import random
from collections.abc import Mapping

from marlben.io.stimulus import Serialized
from marlben.entity import Player
from marlben.entity.npc import NPC
from marlben.io.action import Attack, Style, Melee, Range, Mage
from marlben.lib import colors


class _IdCounter:
    def __init__(self):
        self.reset()

    def reset(self):
        self.npc_id = 0
        self.player_id = 0

    def next_npc_id(self):
        self.npc_id -= 1
        return self.npc_id

    def next_player_id(self):
        self.player_id += 1
        return self.player_id


class GroupsManager:
    def __init__(self, config, realm):
        self.realm = realm
        self.config = config
        self.id_counter = _IdCounter()

        self.npc_groups = [
            NPCGroup(config, realm, npc_group_config, self.id_counter)
            for npc_group_config in self.config.NPC_GROUPS
        ]
        self.player_groups = [
            PlayerGroup(config, realm, player_group_config, self.id_counter, i)
            for i, player_group_config in enumerate(self.config.PLAYER_GROUPS)
        ]

    def spawn(self):
        for npc_group in self.npc_groups:
            npc_group.spawn()
        for player_group in self.player_groups:
            player_group.spawn()

    def reset(self):
        self.id_counter.reset()
        for npc_group in self.npc_groups:
            npc_group.reset()
        for player_group in self.player_groups:
            player_group.reset()

    def get_npc_actions(self):
        actions = {}
        for npc_group in self.npc_groups:
            actions.update(npc_group.actions())
        return actions

    def get_entity_by_id(self, id):
        if id < 0:
            for npc_group in self.npc_groups:
                if id in npc_group.entities:
                    return npc_group[id]
        else:
            for player_group in self.player_groups:
                if id in player_group.entities:
                    return player_group[id]
        return None

    @property
    def packet(self):
        npc_packet = [npc_group.packet for npc_group in self.npc_groups]
        player_packet = [
            player_group.packet for player_group in self.player_groups]
        return self.merge_packets(npc_packet), self.merge_packets(player_packet)

    def merge_packets(self, packets):
        result = {}
        for packet in packets:
            result.update(packet)
        return result

    def update(self, actions):
        for npc_group in self.npc_groups:
            npc_group.update(actions)
        for player_group in self.player_groups:
            player_group.update(actions)

    def update_diary(self):
        for player_group in self.player_groups:
            player_group.update_diary()

    def players_count(self):
        count = 0
        for player_group in self.player_groups:
            count += len(player_group.entities)
        return count

    def mask_player_actions(self, actions: dict):
        result = {}
        for i, action in actions.items():
            for player_group in self.player_groups:
                if i in player_group.entities:
                    result[i] = player_group.mask_action(action)
                    break
        return result

    def cull(self):
        dead = {}
        for npc_group in self.npc_groups:
            npc_group.cull()
        for player_group in self.player_groups:
            dead.update(player_group.cull())
        return dead


class EntityGroup(Mapping):
    def __init__(self, config, realm):
        self.dataframe = realm.dataframe
        self.config = config

        self.entities = {}
        self.dead = {}

    def __len__(self):
        return len(self.entities)

    def __contains__(self, e):
        return e in self.entities

    def __getitem__(self, key):
        return self.entities[key]

    def __iter__(self):
        yield from self.entities

    def items(self):
        return self.entities.items()

    @property
    def corporeal(self):
        return {**self.entities, **self.dead}

    @property
    def packet(self):
        return {k: v.packet() for k, v in self.corporeal.items()}

    def reset(self):
        for entID, ent in self.entities.items():
            self.dataframe.remove(Serialized.Entity, entID, ent.pos)

        self.spawned = False
        self.entities = {}
        self.dead = {}

    def spawn(self, entity):
        pos, entID = entity.pos, entity.entID
        self.realm.map.tiles[pos].addEnt(entity)
        self.entities[entID] = entity

    def cull(self):
        self.dead = {}
        for entID in list(self.entities):
            player = self.entities[entID]
            if not player.alive:
                r, c = player.base.pos
                entID = player.entID
                self.dead[entID] = player

                self.realm.map.tiles[r, c].delEnt(entID)
                del self.entities[entID]
                self.realm.dataframe.remove(
                    Serialized.Entity, entID, player.pos)

        return self.dead

    def update(self, actions):
        for entID, entity in self.entities.items():
            entity.update(self.realm, actions)


class NPCGroup(EntityGroup):
    def __init__(self, config, realm, group_config, id_counter):
        super().__init__(config, realm)
        self.realm = realm
        self.group_config = group_config
        self.coordinate_sampler = self.group_config.SPAWN_COORDINATES_SAMPLER
        self.skills_sampler = self.group_config.SPAWN_SKILLS_SAMPLER
        self.id_counter = id_counter
        self.available_styles = list({Melee, Range, Mage}.difference(
            set(group_config.BANNED_ATTACK_STYLES)))

    def spawn(self):
        if not self.config.game_system_enabled('NPC'):
            return  # TODO: Move to global manager

        for _ in range(self.group_config.NENT - len(self.entities)):
            for _ in range(self.group_config.SPAWN_ATTEMPTS_PER_ENT):
                r, c = self.coordinate_sampler.get_next()
                if self.realm.map.tiles[r, c].occupied:
                    continue

                skills = self.skills_sampler.get_next((r, c))
                # TODO: Check & change
                npc = NPC.spawn(self.realm, (r, c),
                                self.id_counter.next_npc_id(), skills, self.group_config.DANGER)
                npc.skills.style = random.choice(self.available_styles)
                if npc:
                    super().spawn(npc)
                    break

    def reset(self):
        super().reset()
        self.coordinate_sampler.reset(self.config)
        self.skills_sampler.reset(self.config)

    def actions(self):
        actions = {}
        for idx, entity in self.entities.items():
            actions[idx] = entity.decide(self.realm)
        return actions


class PlayerGroup(EntityGroup):
    def __init__(self, config, realm, group_config, id_counter, group_id):
        super().__init__(config, realm)
        self.group_config = group_config
        self.loader = group_config.AGENT_LOADER
        self.palette = colors.Palette()
        self.coordinate_sampler = self.group_config.SPAWN_COORDINATES_SAMPLER
        self.skills_sampler = self.group_config.SPAWN_SKILLS_SAMPLER
        self.banned_attack_styles = self.group_config.BANNED_ATTACK_STYLES
        self.visible_colors = set(list(self.group_config.VISIBLE_COLORS) + [0])
        self.accessible_colors = set(
            list(self.group_config.ACCESSIBLE_COLORS) + [0])
        self.realm = realm
        self.id_counter = id_counter
        self.group_id = group_id

    def spawn(self):
        if self.spawned:
            return
        self.spawned = True
        for _ in range(self.group_config.NENT - len(self.entities)):
            r_f, c_f = None, None
            for _ in range(self.group_config.SPAWN_ATTEMPTS_PER_ENT):
                r, c = self.coordinate_sampler.get_next()
                if not self.realm.map.tiles[r, c].occupied and not self.realm.map.tiles[r, c].impassible:
                    r_f, c_f = r, c
                    break

            if r_f is not None:
                pop_id, agent = next(self.agents)
                agent = agent(self.config, self.id_counter.next_player_id())
                skills = self.skills_sampler.get_next((r_f, c_f))

                player = Player(
                    self.realm, (r_f, c_f), agent,
                    self.palette.color(self.group_id),
                    self.group_id, skills,
                    visible_colors=self.visible_colors,
                    accessible_colors=self.accessible_colors
                )
                super().spawn(player)

    def update_diary(self):
        for entID, entity in self.entities.items():
            entity.update_diary(self.realm)

    def reset(self):
        super().reset()
        self.agents = self.loader(self.group_config)
        self.coordinate_sampler.reset(self.config)
        self.skills_sampler.reset(self.config)

    def mask_action(self, action: dict):
        # action = copy.deepcopy(action)
        if Attack in action:
            if action[Attack][Style] in self.banned_attack_styles:
                action.pop(Attack)
        return action



from pdb import set_trace as T
import numpy as np

import marlben
from marlben.systems.ai import move, attack, utils


def update(entity):
    '''Update validity of tracked entities'''
    if not utils.validTarget(entity, entity.attacker, entity.vision):
        entity.attacker = None
    if not utils.validTarget(entity, entity.target, entity.vision):
        entity.target = None
    if not utils.validTarget(entity, entity.closest, entity.vision):
        entity.closest = None

    if entity.__class__.__name__ != 'Player':
        return

    if not utils.validResource(entity, entity.food, entity.vision):
        entity.food = None
    if not utils.validResource(entity, entity.water, entity.vision):
        entity.water = None


def pathfind(realm, actions, entity, target):
    actions[marlben.action.Move] = {
        marlben.action.Direction: move.pathfind(realm.map.tiles, entity, target)}


def explore(realm, actions, entity):
    sz = realm.config.TERRAIN_SIZE
    r, c = entity.pos

    spawnR, spawnC = entity.spawnPos
    centR, centC = sz//2, sz//2

    vR, vC = centR-spawnR, centC-spawnC

    mmag = max(abs(vR), abs(vC))
    rr = r + int(np.round(entity.vision*vR/mmag))
    cc = c + int(np.round(entity.vision*vC/mmag))

    tile = realm.map.tiles[rr, cc]
    pathfind(realm, actions, entity, tile)


def explore(config, ob, actions, spawnR, spawnC):
    vision = config.NSTIM
    sz = config.TERRAIN_SIZE
    Entity = marlben.Serialized.Entity
    Tile = marlben.Serialized.Tile

    agent = ob.agent
    r = utils.Observation.attribute(agent, Entity.R)
    c = utils.Observation.attribute(agent, Entity.C)

    centR, centC = sz//2, sz//2

    vR, vC = centR-spawnR, centC-spawnC

    mmag = max(abs(vR), abs(vC))
    rr = int(np.round(vision*vR/mmag))
    cc = int(np.round(vision*vC/mmag))

    pathfind(config, ob, actions, rr, cc)


def meander(realm, actions, entity):
    actions[marlben.action.Move] = {
        marlben.action.Direction: move.habitable(realm.map.tiles, entity)}


def evade(realm, actions, entity):
    actions[marlben.action.Move] = {marlben.action.Direction: move.antipathfind(
        realm.map.tiles, entity, entity.attacker)}


def hunt(realm, actions, entity):
    # Move args
    distance = utils.distance(entity, entity.target)

    direction = None
    if distance == 0:
        direction = move.random()
    elif distance > 1:
        direction = move.pathfind(realm.map.tiles, entity, entity.target)

    if direction is not None:
        actions[marlben.action.Move] = {marlben.action.Direction: direction}

    attack(realm, actions, entity)


def attack(realm, actions, entity):
    distance = utils.distance(entity, entity.target)
    if distance > entity.skills.style.attackRange(realm.config):
        return

    actions[marlben.action.Attack] = {
        marlben.action.Style: entity.skills.style,
        marlben.action.Target: entity.target}



import marlben

"""
The base environment, implemented according to the single structure described in https://github.com/jbr-ai-labs/marlben/wiki#marlben. 
Suitable for testing and debugging algorithms before experiments in more complex environments. 
In addition, due to the sparse reward and random map generation, this environment is not trivial, and in itself may be suitable for comparing algorithms.

Expected behavior: 
All agents coordinate and share the resources available on the map, as a result of which they consistently survive until the end of the episode. 
In the event of a shortage of resources for all, there will be competition for resources and only a part of the agents will survive.

Implementation details are located in config file
"""
class Gathering(marlben.Env):
    pass



from marlben.config.base.presets import Medium, Small, Large
from marlben.config.systems.presets import AllGameSystems


class DefaultSmall(Small, AllGameSystems):
    pass


class Default(Medium, AllGameSystems):
    pass



import marlben
from marlben.entity import entity
from marlben.systems.achievement import Diary


class Player(entity.Entity):
    def __init__(self, realm, pos, agent, color, pop, skills, visible_colors, accessible_colors):
        super().__init__(realm, pos, agent.iden, agent.name, color, pop, skills)

        self.agent = agent
        self.pop = pop

        # Scripted hooks
        self.target = None
        self.food = None
        self.water = None
        self.vision = 7

        # Submodules
        self.diary = None
        if tasks := realm.config.TASKS:
            self.diary = Diary(tasks)

        self.visible_colors = visible_colors
        self.accessible_colors = accessible_colors

        self.dataframe.init(marlben.Serialized.Entity, self.entID, self.pos)

    @property
    def serial(self):
        return self.population, self.entID

    @property
    def isPlayer(self) -> bool:
        return True

    @property
    def population(self):
        return self.pop

    def applyDamage(self, dmg, style, stealing_enabled=True):
        if dmg > 0 and stealing_enabled:
            self.resources.food.increment(dmg)
            self.resources.water.increment(dmg)
        self.skills.applyDamage(dmg, style)

    def receiveDamage(self, source, dmg, stealing_enabled=True):
        if not super().receiveDamage(source, dmg, stealing_enabled):
            if source:
                source.history.playerKills += 1
            return
        if dmg > 0 and stealing_enabled:
            self.resources.food.decrement(dmg)
            self.resources.water.decrement(dmg)
        self.skills.receiveDamage(dmg, stealing_enabled)

    def receiveLoot(self, loadout):
        if loadout.chestplate.level > self.loadout.chestplate.level:
            self.loadout.chestplate = loadout.chestplate
        if loadout.platelegs.level > self.loadout.platelegs.level:
            self.loadout.platelegs = loadout.platelegs

    def packet(self):
        data = super().packet()

        data['entID'] = self.entID
        data['annID'] = self.population

        data['base'] = self.base.packet()
        data['resource'] = self.resources.packet()
        data['skills'] = self.skills.packet()

        return data

    def update(self, realm, actions):
        """Post-action update. Do not include history"""
        super().update(realm, actions)

        if not self.alive:
            return

        self.resources.update(realm, self, actions)
        self.skills.update(realm, self, actions)

    def update_diary(self, realm):
        if self.diary:
            self.diary.update(realm, self)



import random
from queue import PriorityQueue

import numpy as np

from marlben.lib import material
from marlben.lib.utils import inBounds
from marlben.scripting import Observation


def validTarget(ent, targ, rng):
    if targ is None or not targ.alive:
        return False
    if lInf(ent, targ) > rng:
        return False
    return True


def validResource(ent, tile, rng):
    return tile is not None and tile.state.tex in (
        'forest', 'water') and distance(ent, tile) <= rng


def directionTowards(ent, targ):
    sr, sc = ent.base.pos
    tr, tc = targ.base.pos

    if abs(sc - tc) > abs(sr - tr):
        direction = (0, np.sign(tc - sc))
    else:
        direction = (np.sign(tr - sr), 0)

    return direction


def closestTarget(ent, tiles, rng=1):
    sr, sc = ent.base.pos
    for d in range(rng + 1):
        for r in range(-d, d + 1):
            for e in tiles[sr + r, sc - d].ents.values():
                if e is not ent and validTarget(ent, e, rng):
                    return e

            for e in tiles[sr + r, sc + d].ents.values():
                if e is not ent and validTarget(ent, e, rng):
                    return e

            for e in tiles[sr - d, sc + r].ents.values():
                if e is not ent and validTarget(ent, e, rng):
                    return e

            for e in tiles[sr + d, sc + r].ents.values():
                if e is not ent and validTarget(ent, e, rng):
                    return e


def distance(ent, targ):
    return l1(ent.pos, targ.pos)


def lInf(ent, targ):
    sr, sc = ent.pos
    gr, gc = targ.pos
    return abs(gr - sr) + abs(gc - sc)


def adjacentPos(pos):
    r, c = pos
    return [(r - 1, c), (r, c - 1), (r + 1, c), (r, c + 1), (r, c)]


def cropTilesAround(position: (int, int), horizon: int, tiles):
    line, column = position

    return tiles[max(line - horizon, 0): min(line + horizon + 1, len(tiles)),
                 max(column - horizon, 0): min(column + horizon + 1, len(tiles[0]))]


def inSight(dr, dc, vision):
    return (
        -vision <= dr <= vision and
        -vision <= dc <= vision)


def vacant(tile):
    from marlben.core import Tile
    occupied = Observation.attribute(tile, Tile.NEnts)
    matl = Observation.attribute(tile, Tile.Index)

    grass = material.Grass.index
    scrub = material.Scrub.index
    forest = material.Forest.index

    return matl in (grass, scrub, forest) and not occupied


def meander(obs):
    cands = []
    if vacant(obs.tile(-1, 0)):
        cands.append((-1, 0))
    if vacant(obs.tile(1, 0)):
        cands.append((1, 0))
    if vacant(obs.tile(0, -1)):
        cands.append((0, -1))
    if vacant(obs.tile(0, 1)):
        cands.append((0, 1))
    if not cands:
        return (-1, 0)
    return random.choice(cands)


# A* Search
def l1(start, goal):
    sr, sc = start
    gr, gc = goal
    return abs(gr - sr) + abs(gc - sc)


def l2(start, goal):
    sr, sc = start
    gr, gc = goal
    return 0.5 * ((gr - sr) ** 2 + (gc - sc) ** 2) ** 0.5


# TODO: unify lInfty and lInf
def lInfty(start, goal):
    sr, sc = start
    gr, gc = goal
    return max(abs(gr - sr), abs(gc - sc))


def aStar(tiles, start, goal, cutoff=100):
    if start == goal:
        return (0, 0)

    pq = PriorityQueue()
    pq.put((0, start))

    backtrace = {}
    cost = {start: 0}

    closestPos = start
    closestHeuristic = l1(start, goal)
    closestCost = closestHeuristic

    while not pq.empty():
        # Use approximate solution if budget exhausted
        cutoff -= 1
        if cutoff <= 0:
            if goal not in backtrace:
                goal = closestPos
            break

        priority, cur = pq.get()

        if cur == goal:
            break

        for nxt in adjacentPos(cur):
            if not inBounds(*nxt, tiles.shape):
                continue
            if tiles[nxt].impassible:
                continue

            newCost = cost[cur] + 1
            if nxt not in cost or newCost < cost[nxt]:
                cost[nxt] = newCost
                heuristic = lInfty(goal, nxt)
                priority = newCost + heuristic

                # Compute approximate solution
                if heuristic < closestHeuristic or (
                        heuristic == closestHeuristic and priority < closestCost):
                    closestPos = nxt
                    closestHeuristic = heuristic
                    closestCost = priority

                pq.put((priority, nxt))
                backtrace[nxt] = cur

    while goal in backtrace and backtrace[goal] != start:
        goal = backtrace[goal]

    sr, sc = start
    gr, gc = goal

    return (gr - sr, gc - sc)


# End A*

# Adjacency functions
def adjacentTiles(tiles, ent):
    r, c = ent.base.pos


def adjacentDeltas():
    return [(-1, 0), (1, 0), (0, 1), (0, -1)]


def l1Deltas(s):
    rets = []
    for r in range(-s, s + 1):
        for c in range(-s, s + 1):
            rets.append((r, c))
    return rets


def posSum(pos1, pos2):
    return pos1[0] + pos2[0], pos1[1] + pos2[1]


def adjacentEmptyPos(env, pos):
    return [p for p in adjacentPos(pos)
            if inBounds(*p, env.size)]


def adjacentTiles(env, pos):
    return [env.tiles[p] for p in adjacentPos(pos)
            if inBounds(*p, env.size)]


def adjacentMats(tiles, pos):
    return [type(tiles[p].state) for p in adjacentPos(pos)
            if inBounds(*p, tiles.shape)]


def adjacentPosWithMat(tiles, pos, mat):
    return [p for p in adjacentPos(pos)
            if inBounds(*p, tiles.shape) and type(tiles[p].state) is mat]


def adjacencyDelMatPairs(env, pos):
    return zip(adjacentDeltas(), adjacentMats(env.tiles, pos))
###End###



from .base_action import Action
from .attack import Attack, Target, Style, Melee, Range, Mage, Heal
from .share import Share, Target, ResourceAmount, Resource, Food, Water
from .move import Move, Direction, West, North, South, East
from .message import Message
from .common import Fixed
from .build import Build, BuildDecision
from .plant import Plant, PlantDecision

Action.hook()



from os import path as osp

from marlben import Agent
from marlben.config import Resource
from marlben.config.base.config import PlayerGroupConfig, Config
from marlben.core.map_generation.pregen_map_generator import PregeneratedMapGenerator
from marlben.envs.gathering.utils.map_generator import GatheringMapGenerator
from marlben.core.spawn.spawn_system.position_samplers import UniformPositionSampler
from marlben.scripted.environments.gathering import GatheringAgent
import math


HORIZON = 20
PATH_TO_CUSTOM_MAPS = osp.dirname(__file__)


class GatheringPlayerGroup(PlayerGroupConfig):

    def __init__(self, n_agents, visible_colors=(), accessible_colors=(), agents=None, coord_sampler=None):
        super().__init__()
        self.NENT = n_agents
        self.VISIBLE_COLORS = visible_colors
        self.SPAWN_ATTEMPTS_PER_ENT = 50
        self.ACCESSIBLE_COLORS = accessible_colors
        self.AGENTS = agents if agents is not None else [Agent]
        if coord_sampler is not None:
            self.SPAWN_COORDINATES_SAMPLER = coord_sampler

def create_group_config(agents, agents_per_group, coord_samplers=None):
    cfgs = []
    if coord_samplers is None:
        coord_samplers = [None] * len(agents)
    for i, agent in enumerate(agents):
        cfg = GatheringPlayerGroup(
            agents=[agent],
            n_agents=agents_per_group,
            coord_sampler=coord_samplers[i]
        )
        cfgs.append(cfg)
    return cfgs


def process_agents(player_groups):
    agents = []
    for group in player_groups:
        agents.extend(group.AGENTS)
    return agents

class BaseGatheringConfig(Resource, Config):
    TRAIN_HORIZON = HORIZON
    EVAL_HORIZON = HORIZON
    MAP_GENERATOR = GatheringMapGenerator
    RESOURCE_COOLDOWN = 8
    RESOURCE_BASE_RESOURCE = 32
    RESOURCE_HARVEST_RESTORE_FRACTION = 1 / RESOURCE_COOLDOWN  # FIXME: Higher cooldown means less resource restoration?
    
    def __init__(self, n_groups, agents_per_group, tiles_per_agent=32, adjust_resource_amount=False):
        super().__init__()
        assert n_groups > 0 and agents_per_group > 0
        total_agents = n_groups * agents_per_group
        map_size = math.ceil((total_agents * tiles_per_agent)**0.5 / 4) * 4  # Round up. Should be divisible by 4
        if adjust_resource_amount:
            # For some environments, it's important to give agent more time to reach resources
            self.RESOURCE_BASE_RESOURCE = 2 * map_size
        self.MAP_WIDTH = map_size
        self.MAP_HEIGHT = map_size
        self.TERRAIN_CENTER = map_size
        self.PATH_MAPS = f'maps/gathering_{map_size}x{map_size}'


class GatheringConfig(BaseGatheringConfig):
    NMAPS = 1
    PLAYER_GROUP_CONFIG = GatheringPlayerGroup

    def __init__(self, n_groups, agents_per_group, tiles_per_agent=32, adjust_resource_amount=False):
        # FIXME: You should not create multiple player groups with the same preset agent
        super().__init__(n_groups, agents_per_group, tiles_per_agent, adjust_resource_amount)
        self.PLAYER_GROUPS = [self.PLAYER_GROUP_CONFIG(agents_per_group, agents=[Agent]) for _ in range(n_groups)]
        self.AGENTS = process_agents(self.PLAYER_GROUPS)


class GatheringConfigScripted(GatheringConfig):
    MAP_GENERATOR = PregeneratedMapGenerator
    RESOURCE_HARVEST_RESTORE_FRACTION = 1.0
    AGENT_TYPE = GatheringAgent

    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = osp.join(PATH_TO_CUSTOM_MAPS, "test_maps")
        for i, group in enumerate(self.PLAYER_GROUPS):
            column = (1 + i // 2) * (1 - i % 2) + (7 - i // 2 - 1) * (i % 2)
            group.SPAWN_COORDINATES_SAMPLER = UniformPositionSampler(r_range=[4, 4], c_range=[column, column])
            group.AGENTS[0] = self.AGENT_TYPE

        self.AGENTS = process_agents(self.PLAYER_GROUPS)



import numpy as np


class Range():
    def __init__(self, mmin, mmax):
        self.mmin = mmin
        self.mmax = mmax

    @property
    def value(self):
        return np.random.randint(self.mmin, self.mmax+1)


class Drop:
    def __init__(self, item, amount, prob):
        self.item = item
        self.amount = amount
        self.prob = prob

    def roll(self):
        if np.random.rand() < self.prob:
            if type(self.amount) == int:
                return (self.item(), self.amount)
            return (self.item(), self.amount.value)


class DropTable:
    def __init__(self):
        self.drops = []

    def add(self, item, quant, prob=1.0):
        self.drops += [Drop(item, quant, prob)]

    def roll(self):
        ret = []
        for e in self.drops:
            drop = e.roll()
            if drop is not None:
                ret += [drop]
        return ret



def inRange(entity, stim, config, N):
    R, C = stim.shape
    R, C = R // 2, C // 2

    rets = set([entity])
    for r in range(R - N, R + N + 1):
        for c in range(C - N, C + N + 1):
            for e in stim[r, c].ents.values():
                rets.add(e)

    rets = list(rets)
    return rets


def l1(start, goal):
    sr, sc = start
    gr, gc = goal
    return abs(gr - sr) + abs(gc - sc)


def l2(start, goal):
    sr, sc = start
    gr, gc = goal
    return 0.5*((gr - sr)**2 + (gc - sc)**2)**0.5



import marlben
from marlben.io.action.common import NodeType, Node
from marlben.io.action.move import Move
from marlben.io.action.attack import Attack
from marlben.io.action.share import Share
from marlben.io.action.build import Build
from marlben.lib.utils import staticproperty


class Action(Node):
    nodeType = NodeType.SELECTION

    @staticproperty
    def edges():
        '''List of valid actions'''
        return [Move, Attack, Share, Build]

    @staticproperty
    def n():
        return len(Action.arguments)

    def args(stim, entity, config):
        return marlben.Serialized.edges

        # Called upon module import (see bottom of file)

    # Sets up serialization domain
    def hook():
        idx = 0
        arguments = []
        for action in Action.edges:
            for args in action.edges:
                if not 'edges' in args.__dict__:
                    continue
                for arg in args.edges:
                    arguments.append(arg)
                    arg.serial = tuple([idx])
                    arg.idx = idx
                    idx += 1
        Action.arguments = arguments



import copy
from dataclasses import dataclass
from typing import Callable


@dataclass
class Task:
    condition: Callable
    target: float = None
    reward: float = 0


class Diary:
    def __init__(self, tasks):
        self.achievements = []
        for task in tasks:
            self.achievements.append(Achievement(
                copy.deepcopy(task.condition), task.target, task.reward))

    @property
    def completed(self):
        return sum(a.completed for a in self.achievements)

    @property
    def cumulative_reward(self, aggregate=True):
        return sum(a.reward * a.completed for a in self.achievements)

    def update(self, realm, entity):
        return {a.name: a.update(realm, entity) for a in self.achievements}


class Achievement:
    def __init__(self, condition, target, reward):
        self.completed = False

        self.condition = condition
        self.target = target
        self.reward = reward

    @property
    def name(self):
        return '{}_{}'.format(self.condition.__name__, self.target)

    def update(self, realm, entity):
        if self.completed:
            return 0

        metric = self.condition(realm, entity)

        if metric >= self.target:
            self.completed = True
            return self.reward

        return 0



import numpy as np

import marlben
from marlben.lib import material


class Tile:
    def __init__(self, config, realm, r, c):
        self.config = config
        self.realm = realm
        self.ents = {}

        self.serialized = 'R{}-C{}'.format(r, c)

        self.r = marlben.Serialized.Tile.R(realm.dataframe, self.serial, r)
        self.c = marlben.Serialized.Tile.C(realm.dataframe, self.serial, c)
        self.nEnts = marlben.Serialized.Tile.NEnts(realm.dataframe, self.serial)
        self.index = marlben.Serialized.Tile.Index(
            realm.dataframe, self.serial, 0)
        self._visibility_color = marlben.Serialized.Tile.VisibilityColor(
            realm.dataframe, self.serial)
        self._accessibility_color = marlben.Serialized.Tile.AccessibilityColor(
            realm.dataframe, self.serial)

        realm.dataframe.init(marlben.Serialized.Tile, self.serial, (r, c))

    @property
    def serial(self):
        return self.serialized

    @property
    def repr(self):
        return ((self.r, self.c))

    @property
    def pos(self):
        return self.r.val, self.c.val

    @property
    def habitable(self):
        return self.mat in material.Habitable

    @property
    def vacant(self):
        return len(self.ents) == 0 and self.habitable

    @property
    def occupied(self):
        return not self.vacant

    @property
    def impassible(self):
        return self.mat in material.Impassible

    @property
    def lava(self):
        return self.mat == material.Lava

    @property
    def static(self):
        """No updates needed"""
        assert self.capacity <= self.mat.capacity
        return self.capacity == self.mat.capacity

    @property
    def visibility_color(self):
        return self._visibility_color.val

    @property
    def accessibility_color(self):
        return self._accessibility_color.val

    def reset(self, mat, vc, ac, config):
        self.state = mat(config)
        self.mat = mat(config)

        self.capacity = self.mat.capacity
        self.current_cooldown = 0
        self.tex = mat.tex
        self.ents = {}
        self._accessibility_color.update(ac)
        self._visibility_color.update(vc)

        self.nEnts.update(0)
        self.index.update(self.state.index)

    def addEnt(self, ent):
        assert ent.entID not in self.ents
        self.nEnts.update(1)
        self.ents[ent.entID] = ent

    def delEnt(self, entID):
        assert entID in self.ents
        self.nEnts.update(0)
        del self.ents[entID]

    def step(self):
        if not self.static:
            if self.config.RESOURCE_COOLDOWN > -1:  # if cooldown system is enabled
                if self.current_cooldown == 0:  # if enough time passed capacity must be restored
                    self.capacity = 1
                else:
                    self.current_cooldown -= 1
            elif np.random.rand() < self.mat.respawn:
                self.capacity += 1
                # FIXME We don't restore original state, do we?

        if self.static:
            self.state = self.mat
            self.index.update(self.state.index)

    def harvest(self):
        if self.capacity == 0:
            return False
        elif self.capacity <= 1:
            # if harvesting, cooldown must be set to max
            self.current_cooldown = self.config.RESOURCE_COOLDOWN
            self.state = self.mat.degen(self.config)
            self.index.update(self.state.index)
        self.capacity -= 1
        return True



import json
import os
import sys
import threading
import time
from signal import signal, SIGINT

import numpy as np
from autobahn.twisted.resource import WebSocketResource
from autobahn.twisted.websocket import WebSocketServerFactory, \
    WebSocketServerProtocol
from twisted.internet import reactor
from twisted.python import log
from twisted.web.server import Site
from twisted.web.static import File


class GodswordServerProtocol(WebSocketServerProtocol):
    def __init__(self):
        super().__init__()
        print("Created a server")
        self.frame = 0

        # "connected" is already used by WSSP
        self.sent_environment = False
        self.isConnected = False

        self.pos = [0, 0]
        self.cmd = None

    def onOpen(self):
        print("Opened connection to server")

    def onClose(self, wasClean, code=None, reason=None):
        self.isConnected = False
        print('Connection closed')

    def connectionMade(self):
        super().connectionMade()
        self.factory.clientConnectionMade(self)

    def connectionLost(self, reason):
        super().connectionLost(reason)
        self.factory.clientConnectionLost(self)
        self.sent_environment = False

    # Not used without player interaction
    def onMessage(self, packet, isBinary):
        print("Server packet", packet)
        packet = packet.decode()
        _, packet = packet.split(';')  # Strip headeer
        r, c, cmd = packet.split(' ')  # Split camera coords
        if len(cmd) == 0 or cmd == '\t':
            cmd = None

        self.pos = [int(r), int(c)]
        self.cmd = cmd

        self.isConnected = True

    def onConnect(self, request):
        print("WebSocket connection request: {}".format(request))
        realm = self.factory.realm
        self.realm = realm
        self.frame += 1

    def serverPacket(self):
        data = self.realm.packet
        return data

    def sendUpdate(self, data):
        packet = {}
        packet['resource'] = data['resource']
        packet['player'] = data['player']
        packet['npc'] = data['npc']
        packet['pos'] = data['pos']
        packet['wilderness'] = data['wilderness']

        config = data['config']

        print('Is Connected? : {}'.format(self.isConnected))
        if not self.sent_environment:
            packet['map'] = data['environment']
            packet['border'] = config.TERRAIN_BORDER
            packet['size'] = config.TERRAIN_SIZE

        if 'overlay' in data:
            packet['overlay'] = data['overlay']
            print('SENDING OVERLAY: ', len(packet['overlay']))

        packet = json.dumps(packet).encode('utf8')
        self.sendMessage(packet, False)


class WSServerFactory(WebSocketServerFactory):
    def __init__(self, ip, realm):
        super().__init__(ip)
        self.realm = realm
        self.time = time.time()
        self.clients = []

        self.pos = [0, 0]
        self.cmd = None
        self.tickRate = 0.6
        self.tick = 0

    def update(self, packet):
        self.tick += 1
        uptime = np.round(self.tickRate * self.tick, 1)
        delta = time.time() - self.time
        print('Wall Clock: ', str(delta)[
              :5], 'Uptime: ', uptime, ', Tick: ', self.tick)
        delta = self.tickRate - delta
        if delta > 0:
            time.sleep(delta)
        self.time = time.time()

        for client in self.clients:
            client.sendUpdate(packet)
            if client.pos is not None:
                self.pos = client.pos
                self.cmd = client.cmd

        return self.pos, self.cmd

    def clientConnectionMade(self, client):
        self.clients.append(client)

    def clientConnectionLost(self, client):
        self.clients.remove(client)


class Application:
    def __init__(self, realm):
        signal(SIGINT, self.kill)
        log.startLogging(sys.stdout)

        port = 8080
        self.factory = WSServerFactory(
            u'ws://localhost:{}'.format(port), realm)
        self.factory.protocol = GodswordServerProtocol
        resource = WebSocketResource(self.factory)

        root = File(".")
        root.putChild(b"ws", resource)
        site = Site(root)

        reactor.listenTCP(port, site)

        def run():
            reactor.run(installSignalHandlers=0)

        threading.Thread(target=run).start()

    def update(self, packet):
        return self.factory.update(packet)

    def kill(*args):
        print("Killed by user")
        reactor.stop()
        os._exit(0)



from marlben.config.base.config import PlayerGroupConfig
from marlben.core.agent import Agent
from marlben.core.spawn.spawn_system.position_samplers import RangePositionSampler
from marlben.core.spawn.spawn_system.skill_samplers import CustomSkillSampler
from marlben.envs.boss_fight.config import BossFightConfig, TankGroupConfig, BossGroupConfig
from marlben.io.action.attack import Range, Heal, Mage, Melee


class FighterGroupConfig(PlayerGroupConfig):
    def __init__(self, n_ent=2, coordinates_sampler=RangePositionSampler(list(range(1, 4)), list(range(3, 7))),
                 skill_sampler=CustomSkillSampler({"constitution": {"name": "const", "level": 50},
                                                   "melee": {"name": "const", "level": 15},
                                                   "range": {"name": "const", "level": 20}})):
        super().__init__()
        self.NENT = n_ent
        self.SPAWN_COORDINATES_SAMPLER = coordinates_sampler
        self.SPAWN_SKILLS_SAMPLER = skill_sampler

    SPAWN_ATTEMPTS_PER_ENT = 10
    BANNED_ATTACK_STYLES = [Heal, Mage]
    AGENTS = [Agent]


class HealerGroupConfig(PlayerGroupConfig):
    def __init__(self, n_ent=2, coordinates_sampler=RangePositionSampler(list(range(1, 4)), list(range(3, 7))),
                 skill_sampler=CustomSkillSampler({"constitution": {"name": "const", "level": 50},
                                                   "heal": {"name": "const", "level": 20}})):
        super().__init__()
        self.NENT = n_ent
        self.SPAWN_COORDINATES_SAMPLER = coordinates_sampler
        self.SPAWN_SKILLS_SAMPLER = skill_sampler

    SPAWN_ATTEMPTS_PER_ENT = 10
    BANNED_ATTACK_STYLES = [Range, Melee, Mage]
    AGENTS = [Agent]


class BossRaidConfig(BossFightConfig):
    def __init__(self, n_tanks=2, n_fighters=2, n_healers=2):
        super().__init__()
        self.NPC_GROUPS = [BossGroupConfig(n_tanks, n_fighters, n_healers)]
        self.PLAYER_GROUPS = [TankGroupConfig(n_tanks), FighterGroupConfig(n_fighters), HealerGroupConfig(n_healers)]



import marlben

"""
Agents are given an additional building action. Its effect is that the cell where the agent was in the previous turn is replaced with a stone cell.

Expected behavior:
Agents can use their already collected resources to trap other agents in parts of the map, or fence themselves off on parts of the map with forest and water tiles.
Thus, the division of the territory between agents can occur.

Implementation details are located in config file
"""
class GatheringBuilding(marlben.Env):
    pass



from collections import defaultdict


class Quill:
    def __init__(self):
        self.blobs = defaultdict(Blob)
        self.stats = defaultdict(list)

    def stat(self, key, val):
        self.stats[key].append(val)

    def register(self, key, tick):
        if key in self.blobs:
            blob = self.blobs[key]
        else:
            blob = Blob()
            self.blobs[key] = blob

        blob.tick = tick
        return blob

    @property
    def packet(self):
        logs = {key: blob.packet for key, blob in self.blobs.items()}
        return {'Log': logs, 'Stats': self.stats}


class Blob:
    def __init__(self):
        self.tracks = defaultdict(Track)

    def log(self, value, key=None):
        self.tracks[key].update(self.tick, value)

    @property
    def packet(self):
        return {key: track.packet for key, track in self.tracks.items()}

# Static blob analytics


class Track:
    def __init__(self):
        self.data = defaultdict(list)

    def update(self, tick, value):
        if type(value) != list:
            value = [value]

        self.data[tick] += value

    @property
    def packet(self):
        return self.data



import marlben

"""
Agents are granted an additional planting action. Its effect is that the cell where the agent was in the previous turn is replaced with a forest cell. 
This action is not free and costs the agent a large amount of food resources. In addition, initially there are fewer forest cells in the environment than in Gathering.

Expected behavior:
Since forest cells are in short supply at the beginning of the episode, all agents can only survive by using the plant action. 
However, since all agents can use the new forest tiles to obtain food, performing this action may be disadvantageous for a particular agent. 
Thus, providing enough food for everyone is an element of the Public Good.

Implementation details are located in config file
"""
class GatheringPlanting(marlben.Env):
    pass



import marlben
from os import path as osp
from .config import BossRaidConfig

"""
A variant of the Boss Fight environment with more agents and more resilient NPCs, requiring a greater degree of coordination as a result.
Besides Tank class, there are also Healer and Fighter classes for agent, which makes the coordination much more difficult since each agent has it's own role and set of actions.

Implementation details are located in config file
"""
class Raid(marlben.Env):
    def __init__(self, config=None):
        if config is None:
            config = BossRaidConfig()
        config.PATH_MAPS = osp.join(osp.abspath(osp.dirname(__file__)), "../../envs/raid/maps")
        super().__init__(config)

    def step(self, actions):
        obs, rewards, dones, infos = super().step(actions)
        if self._done():
            for k in dones:
                dones[k] = True
        return obs, rewards, dones, infos

    def _done(self):
        return len(self.realm.entity_group_manager.npc_groups[0].dead) > 0



from marlben import MapGenerator, Terrain
from marlben.core.map_generation.rock_patterns import generate_huge_rocks
from marlben.core.map_generation.resources import generate_resources
from marlben.core.map_generation.checks import check_connectivity
import numpy as np

from marlben.lib import material


class GatheringMapGenerator(MapGenerator):
    def __init__(self, config):
        super().__init__(config)
        self._seed = 0
        self._proba_decay_coef = 0.9
        self._start_rock_proba = 0.1
        self._start_growth_proba = 0.1
        self._curr_rock_proba = None
        self._curr_growth_proba = None

    def generate_map(self, idx):
        # Using prime numbers to avoid seed repetition
        seed = (271 * self._seed + idx * 193) % 479001599
        self._curr_growth_proba = self._start_growth_proba
        self._curr_rock_proba = self._start_rock_proba
        map = self._generate_with_seed(seed)
        st_pts = np.argwhere(map[:, :, 0] == Terrain.GRASS)
        np.random.shuffle(st_pts)
        while not check_connectivity(map, st_pts[:3], "FULL"):
            seed += 117
            self._curr_growth_proba *= self._proba_decay_coef
            self._curr_rock_proba *= self._proba_decay_coef
            map = self._generate_with_seed(seed)
            st_pts = np.argwhere(map[:, :, 0] == Terrain.GRASS)
            np.random.shuffle(st_pts)

        val = np.zeros((map.shape[0], map.shape[1]))
        val[map[:, :, 0] == Terrain.STONE] = np.random.rand(
        ) * (1 - self.config.TERRAIN_FOREST) + self.config.TERRAIN_FOREST
        val[map[:, :, 0] == Terrain.FOREST] = np.random.rand(
        ) * (self.config.TERRAIN_FOREST - self.config.TERRAIN_GRASS) + self.config.TERRAIN_GRASS
        val[map[:, :, 0] == Terrain.GRASS] = np.random.rand(
        ) * (self.config.TERRAIN_GRASS - self.config.TERRAIN_WATER) + self.config.TERRAIN_WATER
        val[map[:, :, 0] == Terrain.WATER] = np.random.rand(
        ) * (self.config.TERRAIN_WATER - self.config.TERRAIN_LAVA) + self.config.TERRAIN_LAVA
        val[map[:, :, 0] == Terrain.LAVA] = self.config.TERRAIN_LAVA

        return val, map

    def seed(self, seed):
        self._seed = seed

    def _generate_with_seed(self, seed):
        config = self.config
        rnd = np.random.RandomState(seed)
        border = config.TERRAIN_BORDER
        width = config.MAP_WIDTH
        height = config.MAP_HEIGHT
        water_count = config.NENT
        food_count = 2*config.NENT
        rock_start_probability = self._curr_rock_proba  # config.ROCK_FREQ
        rock_growth_probability = self._curr_growth_proba  # config.ROCK_GROWTH
        privacy_colors_num = config.NUM_VISIBILITY_COLORS
        access_colors_num = config.NUM_ACCESSIBILITY_COLORS

        assert config.MAP_HEIGHT * config.MAP_WIDTH >= 6 * config.NENT

        tiles = np.zeros((height + 2*border, width + 2*border, 3), dtype=int)
        tiles[:, :, 0] = Terrain.LAVA
        map = tiles[border:height+border, border:width+border]
        map[:, :, 0] = Terrain.GRASS
        rocks = generate_huge_rocks(height, width, rock_start_probability, rock_growth_probability,
                                    max(1, (width + height) // 8), rnd)
        map[:, :, 0][rocks] = Terrain.STONE
        generate_resources(map, material.BalancedWater.index, water_count,
                           privacy_colors_num, access_colors_num, "EQUAL", "EQUAL", rnd)
        generate_resources(map, material.BalancedForest.index, food_count,
                           privacy_colors_num, access_colors_num, "EQUAL", "EQUAL", rnd)

        return tiles



from marlben.envs import GatheringConfig


class ObscuredAndExclusiveGatheringConfig(GatheringConfig):
    def __init__(self, n_groups, agents_per_group, tiles_per_agent=32, adjust_resource_amount=False):
        super().__init__(n_groups, agents_per_group, tiles_per_agent, adjust_resource_amount)
        self.NUM_ACCESSIBILITY_COLORS = n_groups
        self.NUM_VISIBILITY_COLORS = n_groups
        self.RESOURCE_HARVEST_RESTORE_FRACTION = 1. - 1 / self.RESOURCE_COOLDOWN

        for i, group in enumerate(self.PLAYER_GROUPS):
            group.ACCESSIBLE_COLORS = [i+1]
            group.VISIBLE_COLORS = [i+1]

        self.PATH_MAPS = f'maps/exploring_{self.MAP_WIDTH}x{self.MAP_HEIGHT}'



from enum import Enum, auto

from marlben.lib import utils
from marlben.lib.utils import staticproperty


class NodeType(Enum):
    # Tree edges
    STATIC = auto()  # Traverses all edges without decisions
    SELECTION = auto()  # Picks an edge to follow

    # Executable actions
    ACTION = auto()  # No arguments
    CONSTANT = auto()  # Constant argument
    VARIABLE = auto()  # Variable argument


class Node(metaclass=utils.IterableNameComparable):
    @staticproperty
    def edges():
        return []

    # Fill these in
    @staticproperty
    def priority():
        return None

    @staticproperty
    def type():
        return None

    @staticproperty
    def leaf():
        return False

    @classmethod
    def N(cls, config):
        return len(cls.edges)

    def args(stim, entity, config):
        return []


class Fixed:
    pass



import marlben

"""
As in the Spying environment, some of the resource cells are hidden from each agent and can only be revealed upon interaction.

An additional complexity is that agents and resources have colors. An agent of a certain color can only interact with resource tiles of the same color.
At the same time, among the resource cells visible to the agent from the beginning of the episode, there may not be cells of the color it needs.

Expected behavior:
The agent can bypass the map until it accidentally stumbles upon resources of the desired color.
But a more stable strategy is to exchange information with agents of other colors. 
That is, to send them information about the visible resources of the color they need in the hope of receiving useful information in return.
Since there is no competition for resources between agents of different colors, such an exchange is indeed mutually beneficial. 
Therefore, the environment is well-suited for testing communication approaches.

Implementation details are located in config file
"""
class GatheringObscuredAndExclusive(marlben.Env):
    pass



from os import path as osp

from marlben.config import Building
from marlben.envs import GatheringConfig, GatheringConfigScripted
from marlben.scripted.environments.gathering import GatheringBuildingAgent

PATH_TO_CUSTOM_MAPS = osp.dirname(__file__)


class BuildingGatheringConfig(GatheringConfig, Building):
    def __init__(self, n_groups, agents_per_group, tiles_per_agent=32, adjust_resource_amount=False):
        super().__init__(n_groups, agents_per_group, tiles_per_agent, adjust_resource_amount)
        self.PATH_MAPS = f'maps/building_{self.MAP_WIDTH}x{self.MAP_HEIGHT}'



class BuildingGatheringConfigScripted(GatheringConfigScripted, Building):
    AGENT_TYPE = GatheringBuildingAgent
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = osp.join(PATH_TO_CUSTOM_MAPS, "test_maps")



import marlben
from os import path as osp

"""
The environment is a narrow corridor of grass cells surrounded by stone cells. 
Agents starting the episode on opposite sides of the map can only move horizontally along this corridor. 
There is also a forest space on one side of the corridor, and a water space on the other.
Agents are also given an additional action - to share the selected resource with other agents if it is in the neighborhood.

Expected behavior:
This environment requires specific coordinated behavior from two agents. 
Since agents cannot move through each other, the only way to have both resources in abundance is to trade with another agent. 
Thus, for mutual survival, both agents must collect the resources available to them at their ends of the corridor and periodically meet in the center of the corridor for an exchange.

Implementation details are located in config file
"""
class Corridor(marlben.Env):
    def __init__(self, config):
        super().__init__(config)



from pdb import set_trace as T
import numpy as np


class ExperienceCalculator:
    def __init__(self):
        self.exp = [0]
        self.tabulateExp()

    def tabulateExp(self, numLevels=2999):
        for i in range(2, numLevels+1):
            increment = np.floor(i-1 + 300*(2**((i-1)/7.0)))/4.0
            self.exp += [self.exp[-1] + increment]

        self.exp = np.floor(np.array(self.exp))

    def expAtLevel(self, level):
        return self.exp[level - 1]

    def levelAtExp(self, exp):
        if exp >= self.exp[-1]:
            return len(self.exp)
        return np.argmin(exp >= self.exp)



import marlben

"""
This environment is a simplified version of the Exploring environment. In it, all information about resource cells is initially available to all agents. 
As a consequence, there is no need to exchange information. This environment is suitable for testing and debugging algorithms before moving on to a more complex Exploring environment.

Expected behavior:
Agents collect resources of their color to survive on the map.

Implementation details are located in config file
"""
class GatheringExclusive(marlben.Env):
    pass



import numpy as np


class Sampler:
    def get_next(self):
        pass

    def reset(self):
        pass


class ConstantSampler(Sampler):
    def __init__(self, c):
        self.c = c

    def get_next(self):
        return self.c


class RangeSampler(Sampler):
    def __init__(self, a, b):
        self.i = 0
        self.a = a
        self.b = b

    def get_next(self):
        x = self.a + (self.b - self.a) * self.i / abs(self.b - self.a)
        self.i = (self.i + 1) % (abs(self.b - self.a) + 1)
        return x

    def reset(self):
        self.i = 0


class UniformSampler(Sampler):
    def __init__(self, a, b):
        self.a = a
        self.b = b

    def get_next(self):
        return np.random.randint(self.a, self.b+1)


class ListSampler(Sampler):
    def __init__(self, list):
        self.list = list
        self.i = 0

    def get_next(self):
        x = self.list[self.i]
        self.i = (self.i + 1) % len(self.list)
        return x

    def reset(self):
        self.i = 0


class ChoiceSampler(Sampler):
    def __init__(self, list):
        self.list = list

    def get_next(self):
        return np.random.choice(self.list)



from dataclasses import dataclass, field

from marlben.config.base.config import Config
from marlben.core.spawn.spawn_system import skill_samplers
from .config import NPCGroupConfig, PlayerGroupConfig
from ... import Agent
from ...core.spawn.spawn_system.skill_samplers import SkillsSampler


class SmallNPCGroupConfig(NPCGroupConfig):
    NENT = 32
    SPAWN_SKILLS_SAMPLER = skill_samplers.DefaultNPCSkillSampler(1, 10, 1)


class SmallPlayerGroupConfig(PlayerGroupConfig):
    NENT = 1
    AGENTS = [Agent]


@dataclass
class Small(Config):
    '''A small config for debugging and experiments with an expensive outer loop'''

    PATH_MAPS: str = 'maps/small'
    MAP_PREVIEW_DOWNSCALE: int = 4

    TERRAIN_LOG_INTERPOLATE_MIN: int = 0

    TERRAIN_CENTER: int = 8  # 4
    MAP_HEIGHT: int = 4
    MAP_WIDTH: int = 4
    PLAYER_GROUPS: list = field(default_factory=lambda: [SmallPlayerGroupConfig(), SmallPlayerGroupConfig()])


@dataclass
class MediumNPCGroupConfig(NPCGroupConfig):
    NENT: int = 128
    SPAWN_SKILLS_SAMPLER: SkillsSampler = skill_samplers.DefaultNPCSkillSampler(1, 30, 5)


@dataclass
class MediumPlayerGroupConfig(PlayerGroupConfig):
    NENT: int = 256


@dataclass
class Medium(Config):
    '''A medium config suitable for most academic-scale research'''

    PATH_MAPS = 'maps/medium'
    MAP_PREVIEW_DOWNSCALE = 16

    TERRAIN_CENTER = 128
    MAP_HEIGHT = 128
    MAP_WIDTH = 128

    PLAYER_GROUPS = [MediumPlayerGroupConfig()]
    NPC_GROUPS = [MediumNPCGroupConfig()]


@dataclass
class LargeNPCGroupConfig(NPCGroupConfig):
    NENT = 1024
    SPAWN_SKILLS_SAMPLER = skill_samplers.DefaultNPCSkillSampler(1, 99, 10)


@dataclass
class LargePlayerGroupConfig(PlayerGroupConfig):
    NENT = 2048


@dataclass
class Large(Config):
    '''A large config suitable for large-scale research or fast models'''

    PATH_MAPS = 'maps/large'
    MAP_PREVIEW_DOWNSCALE = 64

    TERRAIN_CENTER = 1024
    MAP_HEIGHT = 1024
    MAP_WIDTH = 1024
    PLAYER_GROUPS = [LargePlayerGroupConfig()]
    NPC_GROUPS = [LargeNPCGroupConfig()]



from marlben.envs import GatheringConfig


class ObscuredGatheringConfig(GatheringConfig):
    def __init__(self, n_groups, agents_per_group, tiles_per_agent=32, adjust_resource_amount=False):
        super().__init__(n_groups, agents_per_group, tiles_per_agent, adjust_resource_amount)
        self.RESOURCE_HARVEST_RESTORE_FRACTION = self.RESOURCE_COOLDOWN / self.RESOURCE_BASE_RESOURCE
        self.NUM_VISIBILITY_COLORS = n_groups
        for i, group in enumerate(self.PLAYER_GROUPS):
            group.VISIBLE_COLORS = [i+1]
        self.PATH_MAPS = f'maps/spying_{self.MAP_WIDTH}x{self.MAP_HEIGHT}'



import marlben
import numpy as np
from marlben import scripting
from marlben.systems.ai import move, attack, utils


def update(entity):
    '''Update validity of tracked entities'''
    if not utils.validTarget(entity, entity.attacker, entity.vision):
        entity.attacker = None
    if not utils.validTarget(entity, entity.target, entity.vision):
        entity.target = None
    if not utils.validTarget(entity, entity.closest, entity.vision):
        entity.closest = None

    if entity.__class__.__name__ != 'Player':
        return

    if not utils.validResource(entity, entity.food, entity.vision):
        entity.food = None
    if not utils.validResource(entity, entity.water, entity.vision):
        entity.water = None


def pathfind(config, ob, actions, rr, cc):
    actions[marlben.action.Move] = {marlben.action.Direction: move.pathfind(config, ob, actions, rr, cc)}


def explore(config, ob, actions, spawnR, spawnC):
    vision = config.NSTIM
    sz = config.TERRAIN_SIZE
    Entity = marlben.Serialized.Entity
    Tile = marlben.Serialized.Tile

    agent = ob.agent
    r = scripting.Observation.attribute(agent, Entity.R)
    c = scripting.Observation.attribute(agent, Entity.C)

    centR, centC = sz // 2, sz // 2

    vR, vC = centR - spawnR, centC - spawnC

    mmag = max(abs(vR), abs(vC))
    rr = int(np.round(vision * vR / mmag))
    cc = int(np.round(vision * vC / mmag))

    pathfind(config, ob, actions, rr, cc)


def meander(realm, actions, entity):
    actions[marlben.action.Move] = {marlben.action.Direction: move.habitable(realm.map.tiles, entity)}


def evade(realm, actions, entity):
    actions[marlben.action.Move] = {marlben.action.Direction: move.antipathfind(realm.map.tiles, entity, entity.attacker)}


def hunt(realm, actions, entity):
    # Move args
    distance = utils.distance(entity, entity.target)

    direction = None
    if distance == 0:
        direction = move.random()
    elif distance > 1:
        direction = move.pathfind(realm.map.tiles, entity, entity.target)

    if direction is not None:
        actions[marlben.action.Move] = {marlben.action.Direction: direction}

    attack(realm, actions, entity)


def attack(realm, actions, entity):
    distance = utils.distance(entity, entity.target)
    if distance > entity.skills.style.attackRange(realm.config):
        return

    actions[marlben.action.Attack] = {marlben.action.Style: entity.skills.style,
                                      marlben.action.Target: entity.target}


def forageDP(realm, actions, entity):
    direction = utils.forageDP(realm.map.tiles, entity)
    actions[marlben.action.Move] = {marlben.action.Direction: move.towards(direction)}


# def forageDijkstra(realm, actions, entity):
def forageDijkstra(config, ob, actions, food_max, water_max):
    direction = utils.forageDijkstra(config, ob, food_max, water_max)
    actions[marlben.action.Move] = {marlben.action.Direction: move.towards(direction)}



import numpy as np

from .base_samplers import UniformSampler, ListSampler, ChoiceSampler


class PositionSampler:
    def __init__(self):
        pass

    def get_next(self):
        pass

    def reset(self, config):
        pass


class ConcurrentPositionSampler(PositionSampler):
    def __init__(self):
        super().__init__()
        self.x_sampler = None
        self.y_sampler = None

    def get_next(self):
        return self.x_sampler.get_next(), self.y_sampler.get_next()

    def reset(self, config):
        map_height = config.MAP_HEIGHT
        map_width = config.MAP_WIDTH
        top, left = config.TERRAIN_BORDER, config.TERRAIN_BORDER

        horizontal_range = np.arange(left, left + map_width, 1).tolist()
        vertical_range = np.arange(top, top + map_height, 1).tolist()

        lefts = (left + np.zeros(len(vertical_range), dtype=int)).tolist()
        rights = ((left + map_width) + np.zeros(len(vertical_range), dtype=int)).tolist()
        highs = (top + np.zeros(len(horizontal_range), dtype=int)).tolist()
        lows = ((top + map_height) + np.zeros(len(horizontal_range), dtype=int)).tolist()

        border_y = horizontal_range + lefts + horizontal_range + rights
        border_x = lows + vertical_range + highs + vertical_range
        zipped = list(zip(border_x, border_y))
        np.random.shuffle(zipped)
        border_x, border_y = zip(*zipped)  # Unzip operation
        self.x_sampler = ListSampler(border_x)
        self.y_sampler = ListSampler(border_y)


class ContinuousPositionSampler(PositionSampler):
    def __init__(self):
        super().__init__()
        self.x_border_choice = None
        self.y_border_choice = None
        self.x_range_sampler = None
        self.y_range_sampler = None

    def get_next(self):
        if np.random.rand() < 0.5:
            return self.x_border_choice.get_next(), self.y_range_sampler.get_next()
        else:
            return self.x_range_sampler.get_next(), self.y_border_choice.get_next()

    def reset(self, config):
        if self.x_border_choice is None:
            map_height = config.MAP_HEIGHT
            map_width = config.MAP_WIDTH
            top, left = config.TERRAIN_BORDER, config.TERRAIN_BORDER
            self.x_border_choice = ChoiceSampler([left, left + map_width])
            self.y_border_choice = ChoiceSampler([top, top + map_height])
            self.x_range_sampler = UniformSampler(left, left + map_width)
            self.y_range_sampler = UniformSampler(top, top + map_height)


class UniformPositionSampler(PositionSampler):
    def __init__(self, r_range=None, c_range=None):
        super().__init__()
        self.x_sampler = None
        self.y_sampler = None
        self.r_range = r_range
        self.c_range = c_range

    def reset(self, config):
        map_height = config.MAP_HEIGHT
        map_width = config.MAP_WIDTH
        top, left = config.TERRAIN_BORDER, config.TERRAIN_BORDER

        if self.r_range is None:
            r_start, r_end = top, top + map_height
        else:
            r_start, r_end = [r + top for r in self.r_range]

        if self.c_range is None:
            c_start, c_end = left, left + map_width
        else:
            c_start, c_end = [c + left for c in self.c_range]

        self.x_sampler = UniformSampler(r_start, r_end)
        self.y_sampler = UniformSampler(c_start, c_end)

    def get_next(self):
        return self.x_sampler.get_next(), self.y_sampler.get_next()


class RangePositionSampler(PositionSampler):
    def __init__(self, r_range, c_range):
        super().__init__()
        self.x_sampler = None
        self.y_sampler = None
        self.r_range = r_range
        self.c_range = c_range

    def get_next(self):
        return self.x_sampler.get_next(), self.y_sampler.get_next()

    def reset(self, config):
        if self.x_sampler is None:
            r_range = [r + config.TERRAIN_BORDER for r in self.r_range]
            c_range = [c + config.TERRAIN_BORDER for c in self.c_range]

            self.x_sampler = ChoiceSampler(r_range)
            self.y_sampler = ChoiceSampler(c_range)



from typing import List

#from forge.blade.core import material
from marlben.systems import ai

import math

import numpy as np


def map_to_rewards(tiles, entity) -> List[List[float]]:
    lava_reward = stone_reward = water_reward = float('-inf')
    forest_reward = 1.0 + math.pow(
        (1 - entity.resources.food.val / entity.resources.food.max) * 15.0,
        1.25)
    scrub_reward = 1.0
    around_water_reward = 1.0 + math.pow(
        (1 - entity.resources.water.val / entity.resources.water.max) * 15.0,
        1.25)

    reward_matrix = np.full((len(tiles), len(tiles[0])), 0.0)

    for line in range(len(tiles)):
        tile_line = tiles[line]
        for column in range(len(tile_line)):
            tile_val = tile_line[column].state.tex
            if tile_val == 'lava':
                reward_matrix[line][column] += lava_reward

            if tile_val == 'stone':
                reward_matrix[line][column] += stone_reward

            if tile_val == 'forest':
                reward_matrix[line][column] += forest_reward

            if tile_val == 'water':
                reward_matrix[line][column] += water_reward

            # TODO: Make these comparisons work off of the water Enum type
            # instead of string compare
            if 'water' in ai.utils.adjacentMats(tiles, (line, column)):
                reward_matrix[line][column] += around_water_reward

            if tile_val == 'scrub':
                reward_matrix[line][column] += scrub_reward

    return reward_matrix


def compute_values(reward_matrix: List[List[float]]) -> List[List[float]]:
    gamma_factor = 0.8  # look ahead ∈ [0, 1]
    max_delta = 0.01  # maximum allowed approximation

    value_matrix = np.full((len(reward_matrix), len(reward_matrix[0])), 0.0)

    delta = float('inf')
    while delta > max_delta:
        old_value_matrix = np.copy(value_matrix)
        for line in range(len(reward_matrix)):
            for column in range(len(reward_matrix[0])):
                reward = reward_matrix[line][column]
                value_matrix[line][
                    column] = reward + gamma_factor * max_value_around(
                    (line, column), value_matrix)

        delta = np.amax(
            np.abs(np.subtract(old_value_matrix, value_matrix)))
    return value_matrix


def values_around(position: (int, int), value_matrix: List[List[float]]) -> (
        float, float, float, float):
    line, column = position

    if line - 1 >= 0:
        top_value = value_matrix[line - 1][column]
    else:
        top_value = float('-inf')

    if line + 1 < len(value_matrix):
        bottom_value = value_matrix[line + 1][column]
    else:
        bottom_value = float('-inf')

    if column - 1 >= 0:
        left_value = value_matrix[line][column - 1]
    else:
        left_value = float('-inf')

    if column + 1 < len(value_matrix[0]):
        right_value = value_matrix[line][column + 1]
    else:
        right_value = float('-inf')

    return top_value, bottom_value, left_value, right_value


def max_value_around(position: (int, int),
                     value_matrix: List[List[float]]) -> float:
    return max(values_around(position, value_matrix))


def max_value_position_around(position: (int, int),
                              value_matrix: List[List[float]]) -> (int, int):
    line, column = position
    top_value, bottom_value, left_value, right_value = values_around(position,
                                                                     value_matrix)

    max_value = max(top_value, bottom_value, left_value, right_value)

    if max_value == top_value:
        return line - 1, column
    elif max_value == bottom_value:
        return line + 1, column
    elif max_value == left_value:
        return line, column - 1
    elif max_value == right_value:
        return line, column + 1


def max_value_direction_around(position: (int, int),
                               value_matrix: List[List[float]]) -> (int, int):
    top_value, bottom_value, left_value, right_value = values_around(position,
                                                                     value_matrix)

    max_value = max(top_value, bottom_value, left_value, right_value)

    if max_value == top_value:
        return -1, 0
    elif max_value == bottom_value:
        return 1, 0
    elif max_value == left_value:
        return 0, -1
    elif max_value == right_value:
        return 0, 1



class Baseline:
    '''Scale arguments for reproducing baselines

   Requires a 32 core 1 GPU machine
   '''
    NUM_GPUS = 1
    NUM_WORKERS = 28
    EVALUATION_NUM_WORKERS = 3
    EVALUATION_NUM_EPISODES = 3


class Debug:
    '''Scale arguments for debugging

   Requires a 32 core 1 GPU machine
   '''

    NUM_GPUS = 0
    NUM_WORKERS = 2
    EVALUATION_NUM_WORKERS = 1
    EVALUATION_NUM_EPISODES = 1



# Various Enums used for handling materials, entity types, etc.
# Data texture pairs are used for enums that require textures.
# These textures are filled in by the Render class at run time.

import colorsys

import numpy as np


def rgb(h):
    h = h.lstrip('#')
    return tuple(int(h[i:i+2], 16) for i in (0, 2, 4))


def rgbNorm(h):
    h = h.lstrip('#')
    return tuple(int(h[i:i+2], 16)/255.0 for i in (0, 2, 4))


def makeColor(idx, h=1, s=1, v=1):
    r, g, b = colorsys.hsv_to_rgb(h, s, v)
    rgbval = tuple(int(255*e) for e in [r, g, b])
    hexval = '%02x%02x%02x' % rgbval
    return Color(str(idx), hexval)


class Color:
    def __init__(self, name, hexVal):
        self.name = name
        self.hex = hexVal
        self.rgb = rgb(hexVal)
        self.norm = rgbNorm(hexVal)
        self.value = self.rgb  # Emulate enum

    def packet(self):
        return self.hex


class Color256:
    def make256():
        parh, parv = np.meshgrid(np.linspace(
            0.075, 1, 16), np.linspace(0.25, 1, 16)[::-1])
        parh, parv = parh.T.ravel(), parv.T.ravel()
        idxs = np.arange(256)
        params = zip(idxs, parh, parv)
        colors = [makeColor(idx, h=h, s=1, v=v) for idx, h, v in params]
        return colors
    colors = make256()


class Color16:
    def make():
        hues = np.linspace(0, 1, 16)
        idxs = np.arange(256)
        params = zip(idxs, hues)
        colors = [makeColor(idx, h=h, s=1, v=1) for idx, h in params]
        return colors
    colors = make()


class Tier:
    BLACK = Color('BLACK', '#000000')
    WOOD = Color('WOOD', '#784d1d')
    BRONZE = Color('BRONZE', '#db4508')
    SILVER = Color('SILVER', '#dedede')
    GOLD = Color('GOLD', '#ffae00')
    PLATINUM = Color('PLATINUM', '#cd75ff')
    DIAMOND = Color('DIAMOND', '#00bbbb')


class Neon():
    RED = Color('RED', '#ff0000')
    ORANGE = Color('ORANGE', '#ff8000')
    YELLOW = Color('YELLOW', '#ffff00')

    GREEN = Color('GREEN', '#00ff00')
    MINT = Color('MINT', '#00ff80')
    CYAN = Color('CYAN', '#00ffff')

    BLUE = Color('BLUE', '#0000ff')
    PURPLE = Color('PURPLE', '#8000ff')
    MAGENTA = Color('MAGENTA', '#ff00ff')

    FUCHSIA = Color('FUCHSIA', '#ff0080')
    SPRING = Color('SPRING', '#80ff80')
    SKY = Color('SKY', '#0080ff')

    WHITE = Color('WHITE', '#ffffff')
    GRAY = Color('GRAY', '#666666')
    BLACK = Color('BLACK', '#000000')

    BLOOD = Color('BLOOD', '#bb0000')
    BROWN = Color('BROWN', '#7a3402')
    GOLD = Color('GOLD', '#eec600')
    SILVER = Color('SILVER', '#b8b8b8')

    TERM = Color('TERM', '#41ff00')
    MASK = Color('MASK', '#d67fff')

    def colors():
        return (
            Neon.CYAN, Neon.MINT, Neon.GREEN,
            Neon.BLUE, Neon.PURPLE, Neon.MAGENTA,
            Neon.FUCHSIA, Neon.SPRING, Neon.SKY,
            Neon.RED, Neon.ORANGE, Neon.YELLOW)


class Solid():
    BLUE = Color('BLUE', '#1f77b4')
    ORANGE = Color('ORANGE', '#ff7f0e')
    GREEN = Color('GREEN', '#2ca02c')

    RED = Color('RED',  '#D62728')
    PURPLE = Color('PURPLE', '#9467bd')
    BROWN = Color('BROWN', '#8c564b')

    PINK = Color('PINK', '#e377c2')
    GREY = Color('GREY', '#7f7f7f')
    CHARTREUSE = Color('CHARTREUSE', '#bcbd22')

    SKY = Color('SKY', '#17becf')

    def colors():
        return (
            Solid.BLUE, Solid.ORANGE, Solid.GREEN,
            Solid.RED, Solid.PURPLE, Solid.BROWN,
            Solid.PINK, Solid.CHARTREUSE, Solid.SKY,
            Solid.GREY)


class Palette:
    def __init__(self, initial_swatch=Neon):
        self.colors = {}
        for idx, color in enumerate(initial_swatch.colors()):
            self.colors[idx] = color

    def color(self, idx):
        if idx in self.colors:
            return self.colors[idx]

        color = makeColor(idx, h=np.random.rand(), s=1, v=1)
        self.colors[idx] = color
        return color



import marlben
from marlben import scripting
from marlben.lib import colors

from marlben.scripted import move, attack


class Scripted(marlben.Agent):
    '''Template class for scripted models.
    You may either subclass directly or mirror the __call__ function'''
    scripted = True
    color = colors.Neon.SKY

    def __init__(self, config, idx):
        '''
        Args:
           config : A forge.blade.core.Config object or subclass object
        '''
        super().__init__(config, idx)
        self.food_max = 0
        self.water_max = 0

        self.spawnR = None
        self.spawnC = None

    @property
    def forage_criterion(self) -> bool:
        '''Return true if low on food or water'''
        min_level = 7
        return self.food <= min_level or self.water <= min_level

    def forage(self):
        '''Min/max food and water using Dijkstra's algorithm'''
        move.forageDijkstra(self.config, self.ob, self.actions,
                            self.food_max, self.water_max)

    def explore(self):
        '''Route away from spawn'''
        move.explore(self.config, self.ob, self.actions,
                     self.spawnR, self.spawnC)

    @property
    def downtime(self):
        '''Return true if agent is not occupied with a high-priority action'''
        return not self.forage_criterion and self.attacker is None

    def evade(self):
        '''Target and path away from an attacker'''
        move.evade(self.config, self.ob, self.actions, self.attacker)
        self.target = self.attacker
        self.targetID = self.attackerID
        self.targetDist = self.attackerDist

    def attack(self):
        '''Attack the current target'''
        if self.target is not None:
            assert self.targetID is not None
            attack.target(self.config, self.actions, self.style, self.targetID)

    def select_combat_style(self):
        '''Select a combat style based on distance from the current target'''
        if self.target is None:
            return

        if self.targetDist <= self.config.COMBAT_MELEE_REACH:
            self.style = marlben.action.Melee
        elif self.targetDist <= self.config.COMBAT_RANGE_REACH:
            self.style = marlben.action.Range
        else:
            self.style = marlben.action.Mage

    def target_weak(self):
        '''Target the nearest agent if it is weak'''
        if self.closest is None:
            return False

        selfLevel = scripting.Observation.attribute(self.ob.agent, marlben.Serialized.Entity.Level)
        targLevel = scripting.Observation.attribute(self.closest, marlben.Serialized.Entity.Level)

        if targLevel <= selfLevel <= 5 or selfLevel >= targLevel + 3:
            self.target = self.closest
            self.targetID = self.closestID
            self.targetDist = self.closestDist

    def scan_agents(self, npc_only=False):
        '''Scan the nearby area for agents'''
        self.closest, self.closestDist = attack.closestTarget(
            self.config, self.ob, npc_only)
        self.attacker, self.attackerDist = attack.attacker(
            self.config, self.ob)

        self.closestID = None
        if self.closest is not None:
            self.closestID = scripting.Observation.attribute(
                self.closest, marlben.Serialized.Entity.ID)

        self.attackerID = None
        if self.attacker is not None:
            self.attackerID = scripting.Observation.attribute(
                self.attacker, marlben.Serialized.Entity.ID)

        self.style = None
        self.target = None
        self.targetID = None
        self.targetDist = None

    def adaptive_control_and_targeting(self, explore=True):
        '''Balanced foraging, evasion, and exploration'''
        self.scan_agents()

        if self.attacker is not None:
            self.evade()
            return

        if self.forage_criterion or not explore:
            self.forage()
        else:
            self.explore()

        self.target_weak()

    def __call__(self, obs):
        '''Process observations and return actions
        Args:
           obs: An observation object from the environment. Unpack with scripting.Observation
        '''
        self.actions = {}

        self.ob = scripting.Observation(self.config, obs)
        agent = self.ob.agent

        self.food = scripting.Observation.attribute(
            agent, marlben.Serialized.Entity.Food)
        self.water = scripting.Observation.attribute(
            agent, marlben.Serialized.Entity.Water)

        if self.food > self.food_max:
            self.food_max = self.food
        if self.water > self.water_max:
            self.water_max = self.water

        if self.spawnR is None:
            self.spawnR = scripting.Observation.attribute(
                agent, marlben.Serialized.Entity.R)
        if self.spawnC is None:
            self.spawnC = scripting.Observation.attribute(
                agent, marlben.Serialized.Entity.C)

        self.currR = round(scripting.Observation.attribute(agent, marlben.Serialized.Entity.R))
        self.currC = round(scripting.Observation.attribute(agent, marlben.Serialized.Entity.C))


class Random(Scripted):
    name = 'Random_'
    '''Moves randomly'''

    def __call__(self, obs):
        super().__call__(obs)

        move.random(self.config, self.ob, self.actions)
        return self.actions
    
class Meander(Scripted):
    name = 'Meander_'
    '''Moves randomly on safe terrain'''

    def __call__(self, obs):
        super().__call__(obs)

        move.meander(self.config, self.ob, self.actions)
        return self.actions


class ForageNoExplore(Scripted):
    '''Forages using Dijkstra's algorithm'''
    name = 'ForageNE_'

    def __call__(self, obs):
        super().__call__(obs)

        self.forage()

        return self.actions


class Forage(Scripted):
    '''Forages using Dijkstra's algorithm and actively explores'''
    name = 'Forage_'

    def __call__(self, obs):
        super().__call__(obs)

        if self.forage_criterion:
            self.forage()
        else:
            self.explore()

        return self.actions


class CombatNoExplore(Scripted):
    '''Forages using Dijkstra's algorithm and fights nearby agents'''
    name = 'CombatNE_'

    def __call__(self, obs):
        super().__call__(obs)

        self.adaptive_control_and_targeting(explore=False)

        self.style = marlben.action.Range
        self.attack()

        return self.actions


class Combat(Scripted):
    '''Forages, fights, and explores'''
    name = 'Combat_'

    def __call__(self, obs):
        super().__call__(obs)

        self.adaptive_control_and_targeting()

        self.style = marlben.action.Range
        self.attack()

        return self.actions


class CombatTribrid(Scripted):
    name = 'CombatTri_'
    '''Forages, fights, and explores.
    Uses a slightly more sophisticated attack routine'''

    def __call__(self, obs):
        super().__call__(obs)

        self.adaptive_control_and_targeting()

        self.select_combat_style()
        self.attack()

        return self.actions



"""Setups the project."""

import pathlib

from setuptools import setup

CWD = pathlib.Path(__file__).absolute().parent


def get_version():
    """Gets the marlben version."""
    path = CWD / "marlben" / "version.py"
    content = path.read_text()

    for line in content.splitlines():
        if line.startswith("__version__"):
            return line.strip().split()[-1].strip().strip('"')
    raise RuntimeError("bad version data in __init__.py")


def get_description():
    """Gets the description from the readme."""
    with open("README.md") as fh:
        long_description = ""
        header_count = 0
        for line in fh:
            if line.startswith("##"):
                header_count += 1
            if header_count < 2:
                long_description += line
            else:
                break
    return long_description


setup(name="marlben", version=get_version(), long_description=get_description(), package_data={'resource': ['marlben/resource/*.png']}, include_package_data=True)



import marlben
from marlben.lib import utils
from marlben.systems import combat, equipment
import copy


class Resources:
    def __init__(self, ent):
        self.health = marlben.Serialized.Entity.Health(ent.dataframe, ent.entID)
        self.water = marlben.Serialized.Entity.Water(ent.dataframe, ent.entID)
        self.food = marlben.Serialized.Entity.Food(ent.dataframe, ent.entID)

        self.health.max = ent.skills.constitution.level
        self.water.max = ent.skills.fishing.level
        self.food.max = ent.skills.hunting.level
        self.health.update(self.health.max)
        self.water.update(self.water.max)
        self.food.update(self.food.max)

    def update(self, realm, entity, actions):
        self.health.max = entity.skills.constitution.level
        self.water.max = entity.skills.fishing.level
        self.food.max = entity.skills.hunting.level

    def packet(self):
        data = {}
        data['health'] = self.health.packet()
        data['food'] = self.food.packet()
        data['water'] = self.water.packet()
        return data


class Status:
    def __init__(self, ent):
        self.config = ent.config
        self.freeze = marlben.Serialized.Entity.Freeze(ent.dataframe, ent.entID)

    def update(self, realm, entity, actions):
        self.freeze.decrement()

    def packet(self):
        data = {}
        data['freeze'] = self.freeze.val
        return data


class History:
    def __init__(self, ent):
        self.actions = None
        self.attack = None

        self.origPos = ent.pos
        self.exploration = 0
        self.playerKills = 0

        self.damage = marlben.Serialized.Entity.Damage(ent.dataframe, ent.entID)
        self.timeAlive = marlben.Serialized.Entity.TimeAlive(
            ent.dataframe, ent.entID)

        self.lastPos = None

    def update(self, realm, entity, actions):
        self.attack = None
        self.actions = actions
        self.damage.update(0)

        exploration = utils.linf(entity.pos, self.origPos)
        self.exploration = max(exploration, self.exploration)

        self.timeAlive.increment()

    def packet(self):
        data = {'damage': self.damage.val, 'timeAlive': self.timeAlive.val}

        if self.attack is not None:
            data['attack'] = self.attack

        return data


class Base:
    def __init__(self, ent, pos, iden, name, color, pop):
        self.name = name + str(iden)
        self.color = color
        r, c = pos

        self.r = marlben.Serialized.Entity.R(ent.dataframe, ent.entID, r)
        self.c = marlben.Serialized.Entity.C(ent.dataframe, ent.entID, c)

        self.population = marlben.Serialized.Entity.Population(
            ent.dataframe, ent.entID, pop)
        self.self = marlben.Serialized.Entity.Self(ent.dataframe, ent.entID, 1)
        self.identity = marlben.Serialized.Entity.ID(
            ent.dataframe, ent.entID, ent.entID)
        self.level = marlben.Serialized.Entity.Level(ent.dataframe, ent.entID, 3)

        ent.dataframe.init(marlben.Serialized.Entity, ent.entID, (r, c))

    def update(self, realm, entity, actions):
        self.level.update(combat.level(entity.skills))

    @property
    def pos(self):
        return self.r.val, self.c.val

    def packet(self):
        data = {'r': self.r.val, 'c': self.c.val, 'name': self.name, 'color': self.color.packet(),
                'population': self.population.val, 'self': self.self.val}

        return data

"""
Represents agent on the map with hp and resources
"""
class Entity:
    def __init__(self, realm, pos, iden, name, color, pop, skills):
        self.dataframe = realm.dataframe
        self.config = realm.config
        self.entID = iden
        self.skills = copy.deepcopy(skills)

        self.repr = None
        self.vision = 5

        self.attacker = None
        self.target = None
        self.closest = None
        self.spawnPos = pos

        self.attackerID = marlben.Serialized.Entity.AttackerID(self.dataframe, self.entID, 0)

        # Submodules
        self.base = Base(self, pos, iden, name, color, pop)
        self.status = Status(self)
        self.history = History(self)
        self.resources = Resources(self)
        self.loadout = equipment.Loadout()
        self.visible_colors = set(range(realm.config.NUM_VISIBILITY_COLORS + 1))
        self.accessible_colors = set(range(realm.config.NUM_ACCESSIBILITY_COLORS + 1))

    def packet(self):
        data = {'status': self.status.packet(), 'history': self.history.packet(), 'loadout': self.loadout.packet(),
                'alive': self.alive}

        return data

    def update(self, realm, actions):
        '''Update occurs after actions, e.g. does not include history'''
        if self.history.damage == 0:
            self.attacker = None
            self.attackerID.update(0)

        self.base.update(realm, self, actions)
        self.status.update(realm, self, actions)
        self.history.update(realm, self, actions)

    def update_diary(self, realm):
        pass

    def receiveDamage(self, source, dmg, stealing_enabled=True):
        self.history.damage.update(dmg)
        self.resources.health.decrement(dmg)

        if not self.alive and source:
            source.receiveLoot(self.loadout)
            return False

        return True

    def receiveLoot(self, loadout):
        pass

    def applyDamage(self, dmg, style, stealing_enabled):
        pass

    @property
    def pos(self):
        return self.base.pos

    @property
    def alive(self):
        if self.resources.health.empty:
            return False

        return True

    @property
    def dead(self):
        return not self.alive

    @property
    def isPlayer(self) -> bool:
        return False

    @property
    def isNPC(self) -> bool:
        return False



import numpy as np

from marlben.lib import overlay
from marlben.lib.colors import Neon
from marlben.systems import combat


class OverlayRegistry:
    def __init__(self, config, realm):
        """Manager class for overlays

        Args:
            config: A Config object
            realm: An environment
        """
        self.initialized = False

        self.config = config
        self.realm = realm

        self.overlays = {
            'counts':     Counts,
            'skills':     Skills,
            'wilderness': Wilderness}

    def init(self):
        for cmd, overlay in self.overlays.items():
            self.overlays[cmd] = overlay(self.config, self.realm)

    def step(self, obs, pos, cmd):
        """Per-tick overlay updates

        Args:
            obs: Observation returned by the environment
            pos: Client camera focus position
            cmd: User command returned by the client
        """
        if not self.initialized:
            self.initialized = True
            self.init()

        self.realm.overlayPos = pos
        for overlay_value in self.overlays.values():
            overlay_value.update(obs)

        if cmd in self.overlays:
            self.overlays[cmd].register(obs)


class Overlay:
    """Define a overlay for visualization in the client

    Overlays are color images of the same size as the game map.
    They are rendered over the environment with transparency and
    can be used to gain insight about agent behaviors."""

    def __init__(self, config, realm, *args):
        """
        Args:
            config: A Config object
            realm: An environment
        """
        self.config = config
        self.realm = realm

        self.size = config.TERRAIN_SIZE
        self.values = np.zeros((self.size, self.size))

    def update(self, obs):
        '''Compute per-tick updates to this overlay. Override per overlay.

        Args:
            obs: Observation returned by the environment
        '''
        pass

    def register(self):
        '''Compute the overlay and register it within realm. Override per overlay.'''
        pass


class Skills(Overlay):
    def __init__(self, config, realm, *args):
        '''Indicates whether agents specialize in foraging or combat'''
        super().__init__(config, realm)
        self.nSkills = 2

        self.values = np.zeros((self.size, self.size, self.nSkills))

    def update(self, obs):
        '''Computes a count-based exploration map by painting
        tiles as agents walk over them'''
        for entID, agent in self.realm.realm.players():
            r, c = agent.base.pos

            skillLvl = (agent.skills.fishing.level +
                        agent.skills.hunting.level)/2.0
            combatLvl = combat.level(agent.skills)

            if skillLvl == 10 and combatLvl == 3:
                continue

            self.values[r, c, 0] = skillLvl
            self.values[r, c, 1] = combatLvl

    def register(self, obs):
        values = np.zeros((self.size, self.size, self.nSkills))
        for idx in range(self.nSkills):
            ary = self.values[:, :, idx]
            vals = ary[ary != 0]
            mean = np.mean(vals)
            std = np.std(vals)
            if std == 0:
                std = 1

            values[:, :, idx] = (ary - mean) / std
            values[ary == 0] = 0

        colors = np.array([Neon.BLUE.rgb, Neon.BLOOD.rgb])
        colorized = np.zeros((self.size, self.size, 3))
        amax = np.argmax(values, -1)

        for idx in range(self.nSkills):
            colorized[amax == idx] = colors[idx] / 255
            colorized[values[:, :, idx] == 0] = 0

        self.realm.register(colorized)


class Counts(Overlay):
    def __init__(self, config, realm, *args):
        super().__init__(config, realm)
        self.values = np.zeros((self.size, self.size, config.NPOP))

    def update(self, obs):
        '''Computes a count-based exploration map by painting
        tiles as agents walk over them'''
        for entID, agent in self.realm.realm.players():
            pop = agent.base.population.val
            r, c = agent.base.pos
            self.values[r, c][pop] += 1

    def register(self, obs):
        colors = self.realm.realm.players.palette.colors
        colors = np.array([colors[pop].rgb
                           for pop in range(self.config.NPOP)])

        colorized = self.values[:, :, :, None] * colors / 255
        colorized = np.sum(colorized, -2)
        countSum = np.sum(self.values[:, :], -1)
        data = overlay.norm(countSum)[..., None]

        countSum[countSum == 0] = 1
        colorized = colorized * data / countSum[..., None]

        self.realm.register(colorized)


class Wilderness(Overlay):
    def init(self):
        """Computes the local wilderness level"""
        data = np.zeros((self.size, self.size))
        for r in range(self.size):
            for c in range(self.size):
                data[r, c] = combat.wilderness(self.config, (r, c))

        self.wildy = overlay.twoTone(data, preprocess='clip', invert=True, periods=5)

    def register(self, obs):
        if not hasattr(self, 'wildy'):
            print('Initializing Wilderness')
            self.init()

        self.realm.register(self.wildy)



import random as rand

import marlben
from marlben.systems.ai import utils


def random():
    return rand.choice(marlben.action.Direction.edges)


def randomSafe(tiles, ent):
    r, c = ent.base.pos
    cands = []
    if not tiles[r-1, c].lava:
        cands.append(marlben.action.North)
    if not tiles[r+1, c].lava:
        cands.append(marlben.action.South)
    if not tiles[r, c-1].lava:
        cands.append(marlben.action.West)
    if not tiles[r, c+1].lava:
        cands.append(marlben.action.East)

    return rand.choice(cands)


def habitable(tiles, ent):
    r, c = ent.base.pos
    cands = []
    if tiles[r-1, c].vacant:
        cands.append(marlben.action.North)
    if tiles[r+1, c].vacant:
        cands.append(marlben.action.South)
    if tiles[r, c-1].vacant:
        cands.append(marlben.action.West)
    if tiles[r, c+1].vacant:
        cands.append(marlben.action.East)

    if len(cands) == 0:
        return marlben.action.North

    return rand.choice(cands)


def towards(direction):
    if direction == (-1, 0):
        return marlben.action.North
    elif direction == (1, 0):
        return marlben.action.South
    elif direction == (0, -1):
        return marlben.action.West
    elif direction == (0, 1):
        return marlben.action.East
    else:
        return random()


def bullrush(ent, targ):
    direction = utils.directionTowards(ent, targ)
    return towards(direction)


def pathfind(tiles, ent, targ):
    direction = utils.aStar(tiles, ent.pos, targ.pos)
    return towards(direction)


def antipathfind(tiles, ent, targ):
    er, ec = ent.pos
    tr, tc = targ.pos
    goal = (2*er - tr, 2*ec-tc)
    direction = utils.aStar(tiles, ent.pos, goal)
    return towards(direction)



from dataclasses import dataclass
from typing import Callable
from os import path as osp

from marlben.core.agent import Agent
from marlben.core.spawn.spawn_system.position_samplers import UniformPositionSampler
from marlben.core.spawn.spawn_system.skill_samplers import CustomSkillSampler
from marlben.systems.achievement import Task
from marlben.scripted.environments.corridor import CorridorAgent
from marlben.core.map_generation.pregen_map_generator import PregeneratedMapGenerator

from marlben.config.base.config import PlayerGroupConfig, Config
from marlben.config.systems import Sharing

HORIZON = 70

PATH_TO_CUSTOM_MAPS = osp.dirname(__file__)


class PlayerDiedTask(Callable):
    __name__ = "Die"

    def __call__(self, realm, entity):
        return 0 if entity.alive else 1


class PlayerSurvivedTask(Callable):
    __name__ = "Survive"

    def __init__(self, horizon):
        super().__init__()
        self.horizon = horizon

    def __call__(self, realm, entity):
        survived = entity.alive and realm.tick >= self.horizon
        return 1 if survived else 0


class TraderGroupConfig(PlayerGroupConfig):
    def __init__(self, agents, n_ent, coordinates_sampler, skill_sampler=CustomSkillSampler({})):
        super().__init__()
        self.NENT = n_ent
        self.SPAWN_COORDINATES_SAMPLER = coordinates_sampler
        self.SPAWN_SKILLS_SAMPLER = skill_sampler
        self.AGENTS = agents

    SPAWN_ATTEMPTS_PER_ENT = 1
    BASE_HEALTH = 5
    REGEN_HEALTH = False


class BaseCorridorConfig(Config, Sharing):
    TASKS = [Task(PlayerDiedTask(), 1, -1.0),
             Task(PlayerSurvivedTask(horizon=HORIZON), 1, 1.0)]

    TERRAIN_CENTER = 14
    TERRAIN_BORDER = 9
    NSTIM = 10
    MAP_HEIGHT = 3
    MAP_WIDTH = 10
    TOP_LEFT_CORNER = [17, 12]

    MAP_PREVIEW_DOWNSCALE = 1
    TERRAIN_LOG_INTERPOLATE_MIN = 0
    PATH_MAPS = osp.join(PATH_TO_CUSTOM_MAPS, "maps")
    MAP_GENERATOR = PregeneratedMapGenerator

    RESOURCE_BASE_RESOURCE = 14
    RESOURCE_HARVEST_RESTORE_FRACTION = 1.0
    RESOURCE_COOLDOWN = 5

    TRAIN_HORIZON = HORIZON
    EVAL_HORIZON = HORIZON


def create_group_config(agents):
    return [
        TraderGroupConfig(
            agents=[agents[0]],
            n_ent=1,
            coordinates_sampler=UniformPositionSampler(
                r_range=[1, 1], c_range=[4, 4])
        ),
        TraderGroupConfig(
            agents=[agents[1]],
            n_ent=1,
            coordinates_sampler=UniformPositionSampler(
                r_range=[1, 1], c_range=[5, 5])
        )
    ]


def process_agents(player_groups):
    agents = []
    for group in player_groups:
        agents.extend(group.AGENTS)
    return agents


@dataclass
class ScriptedCorridorConfig(BaseCorridorConfig):
    PLAYER_GROUPS = create_group_config(agents=[CorridorAgent, CorridorAgent])
    AGENTS = process_agents(PLAYER_GROUPS)


class OneNeuralCorridorConfig(BaseCorridorConfig):
    PLAYER_GROUPS = create_group_config(agents=[Agent, CorridorAgent])
    AGENTS = process_agents(PLAYER_GROUPS)


@dataclass
class CorridorConfig(BaseCorridorConfig):
    PLAYER_GROUPS = create_group_config(agents=[Agent, Agent])
    AGENTS = process_agents(PLAYER_GROUPS)



import marlben
from marlben.scripted import move
from marlben.scripted.baselines  import Scripted


class CorridorAgent(Scripted):
    name = 'CorridorAgent_'
    '''Walks along the corridor and shares resources with another agent'''

    def __init__(self, config, idx):
        super().__init__(config, idx)
        self._current_target = "resource"
        self._resource_amount = 7
        if (self.iden + 1) % 2 == 0:
            self._direction = (0, -1)
            self._resource_pos = (10, 10) # (18, 13)
            self._mid_pos = (10, 13)
            self._resource_to_share = marlben.action.Water
        else:
            self._direction = (0, 1)
            self._resource_pos = (10, 17)
            self._mid_pos = (10, 14)
            self._resource_to_share = marlben.action.Food

    def _switch_direction(self, direction):
        return tuple([-1 * d for d in direction])

    def _manage_direction(self, r, c):
        if self._current_target == "resource" and (r, c) == self._resource_pos:
            self._direction = self._switch_direction(self._direction)
            self._current_target = "mid"
        elif self._current_target == "mid" and (r, c) == self._mid_pos:
            self._direction = self._switch_direction(self._direction)
            self._current_target = "resource"

    def _share(self):
        self.scan_agents()
        targetID = self.closestID
        if targetID is None:
            return None
        else:
            return {
                marlben.action.Resource: self._resource_to_share,
                marlben.action.Target: targetID,
                marlben.action.ResourceAmount: self._resource_amount
            }

    def __call__(self, obs):
        super().__call__(obs)

        agent = self.ob.agent
        Entity = marlben.Serialized.Entity
        Tile = marlben.Serialized.Tile

        r = marlben.scripting.Observation.attribute(agent, Entity.R)
        c = marlben.scripting.Observation.attribute(agent, Entity.C)
        self._manage_direction(r, c)
        direction = move.towards(self._direction)
        self.actions[marlben.action.Move] = {marlben.action.Direction: direction}
        if (r, c) == self._mid_pos:
            share_action = self._share()
            if share_action is not None:
                self.actions[marlben.action.Share] = share_action

        return self.actions



import os
from dataclasses import field

import marlben
from marlben.config.common import Template, SequentialLoader
from marlben.core.spawn.spawn_system import position_samplers, skill_samplers
from marlben.io.action import Heal


class NPCGroupConfig(Template):
    SPAWN_COORDINATES_SAMPLER = position_samplers.UniformPositionSampler()
    SPAWN_SKILLS_SAMPLER = skill_samplers.DefaultNPCSkillSampler(1, None, None)
    SPAWN_ATTEMPTS_PER_ENT = 5
    BANNED_ATTACK_STYLES = [Heal]  # By default, monsters can't heal someone
    NENT = 16
    DANGER = None  # Compute vanilla NMMO danger by default


class PlayerGroupConfig(Template):
    AGENT_LOADER = SequentialLoader
    SPAWN_COORDINATES_SAMPLER = position_samplers.ConcurrentPositionSampler()
    SPAWN_SKILLS_SAMPLER = skill_samplers.DefaultSkillSampler()
    SPAWN_ATTEMPTS_PER_ENT = 10
    BANNED_ATTACK_STYLES = []
    VISIBLE_COLORS = []
    ACCESSIBLE_COLORS = []
    AGENTS = []
    NENT = 2


class Config(Template):
    '''An environment configuration object

   Global constants are defined as static class variables. You can override
   any Config variable using standard CLI syntax (e.g. --NENT=128).

   The default config as of v1.5 uses 1024x1024 maps with up to 2048 agents
   and 1024 NPCs. It is suitable to time horizons of 8192+ steps. For smaller
   experiments, consider the SmallMaps config.

   Notes:
      We use Google Fire internally to replace standard manual argparse
      definitions for each Config property. This means you can subclass
      Config to add new static attributes -- CLI definitions will be
      generated automatically.
   '''

    def __init__(self, dictionary=None):
        super().__init__()

        if dictionary is not None:
            self._populate_config(dictionary)

    def _populate_config(self, dictionary):
        for k, v in dictionary.items():
            print(k, v)
            if isinstance(v, dict):
                self._populate_config(v)
            else:
                setattr(self, k, v)
    ############################################################################
    # Meta-Parameters
    RENDER: bool = False

    def items(self):
        class_vars = {}
        for cls in self.__class__.mro()[:-1]:
            class_vars.update({
                k: v for k, v in cls.__dict__.items()
                if k not in object.__dict__ and not k.startswith('_')
                and not callable(v)
                and not isinstance(v, property)
            })
        return class_vars.items()

    def game_system_enabled(self, name) -> bool:
        return hasattr(self, name)

    ############################################################################
    # Visibility and Accessibility settings
    NUM_VISIBILITY_COLORS: int = 0
    NUM_ACCESSIBILITY_COLORS: int = 0

    # Population Parameters
    AGENT_LOADER = SequentialLoader
    '''Agent loader class specifying spawn sampling'''

    TASKS = []
    '''Tasks for which to compute rewards'''

    NPC_GROUPS = []
    PLAYER_GROUPS = []
    '''Groups of players and NPCs to spawn in the environment'''

    NMAPS = 1
    '''Number of maps to generate'''

    NTILE = 16
    # TODO: Find a way to auto-compute this
    '''Number of distinct terrain tile types'''

    NSTIM = 7
    '''Number of tiles an agent can see in any direction'''

    N_AGENT_OBS = 100
    '''Number of distinct agent observations'''

    def WINDOW(self):
        '''Size of the square tile crop visible to an agent'''
        return 2 * self.NSTIM + 1

    ############################################################################
    # Agent Parameters
    BASE_HEALTH = 10
    '''Initial Constitution level and agent health'''

    ############################################################################
    # Terrain Generation Parameters
    MAP_GENERATOR = None
    '''Specifies a user map generator. Uses default generator if unspecified.'''

    FORCE_MAP_GENERATION = True
    '''Whether to regenerate and overwrite existing maps'''

    GENERATE_MAP_PREVIEWS = False
    '''Whether map generation should also save .png previews (slow + large file size)'''

    MAP_PREVIEW_DOWNSCALE = 1
    '''Downscaling factor for png previews'''

    TERRAIN_CENTER = None  # TODO: Remove as we support HEIGHT and WIDTH parameters
    '''Size of each map (number of tiles along each side)'''

    TERRAIN_BORDER = 8
    TOP_LEFT_CORNER = (8, 8)  # TODO: Should implement as a property instead
    '''Number of lava border tiles surrounding each side of the map'''

    @property
    def TERRAIN_SIZE(self):
        return int(self.TERRAIN_CENTER + 2 * self.TERRAIN_BORDER)

    @property
    def NENT(self):
        return sum([cfg.NENT for cfg in self.PLAYER_GROUPS])

    @property
    def NPOP(self):
        return len(self.PLAYER_GROUPS)

    TERRAIN_FLIP_SEED = False
    '''Whether to negate the seed used for generation (useful for unique heldout maps)'''

    TERRAIN_FREQUENCY = -3
    '''Base noise frequency range (log2 space)'''

    TERRAIN_FREQUENCY_OFFSET = 7
    '''Noise frequency octave offset (log2 space)'''

    TERRAIN_LOG_INTERPOLATE_MIN = -2
    '''Minimum interpolation log-strength for noise frequencies'''

    TERRAIN_LOG_INTERPOLATE_MAX = 0
    '''Maximum interpolation log-strength for noise frequencies'''

    TERRAIN_TILES_PER_OCTAVE = 4
    '''Number of octaves sampled from log2 spaced TERRAIN_FREQUENCY range'''

    TERRAIN_LAVA = 0.0
    '''Noise threshold for lava generation'''

    TERRAIN_WATER = 0.30
    '''Noise threshold for water generation'''

    TERRAIN_GRASS = 0.70
    '''Noise threshold for grass'''

    TERRAIN_FOREST = 0.85
    '''Noise threshold for forest'''

    ############################################################################
    # Path Parameters
    PATH_ROOT = os.path.dirname(marlben.__file__)
    '''Global repository directory'''

    PATH_CWD = os.getcwd()
    '''Working directory'''

    PATH_RESOURCE = os.path.join(PATH_ROOT, 'resource')
    '''Resource directory'''

    PATH_TILE = os.path.join(PATH_RESOURCE, '{}.png')
    '''Tile path -- format me with tile name'''

    PATH_MAPS = None
    '''Generated map directory'''

    PATH_MAP_SUFFIX = 'map{}/map.npy'
    '''Map file name'''



class Material:
    harvestable = False
    capacity = 1

    def __init__(self, config):
        pass

    def __eq__(self, mtl):
        return self.index == mtl.index

    def __equals__(self, mtl):
        return self == mtl


class Lava(Material):
    tex = 'lava'
    index = 0


class Water(Material):
    tex = 'water'
    index = 7


class Grass(Material):
    tex = 'grass'
    index = 2


class Scrub(Material):
    tex = 'scrub'
    index = 6


class Forest(Material):
    tex = 'forest'
    index = 8

    harvestable = True
    degen = Scrub

    def __init__(self, config):
        if config.game_system_enabled('Resource'):
            self.capacity = config.RESOURCE_FOREST_CAPACITY
            self.respawn = config.RESOURCE_FOREST_RESPAWN


class Stone(Material):
    tex = 'stone'
    index = 5


class ScrubImpassible(Scrub):
    index = 3


class BalancedWater(Forest):
    degen = ScrubImpassible
    tex = 'water'
    index = 1


class BalancedForest(Forest):
    degen = ScrubImpassible
    tex = 'forest'
    index = 4


class Meta(type):
    def __init__(self, name, bases, dict):
        self.indices = {mtl.index for mtl in self.materials}

    def __iter__(self):
        yield from self.materials

    def __contains__(self, mtl):
        if isinstance(mtl, Material):
            mtl = type(mtl)
        if isinstance(mtl, type):
            return mtl in self.materials
        return mtl in self.indices


class All(metaclass=Meta):
    '''List of all materials'''
    materials = {Lava, Water, Grass, Scrub, Forest, Stone,
                 ScrubImpassible, BalancedForest, BalancedWater}


class Impassible(metaclass=Meta):
    '''Materials that agents cannot walk through'''
    materials = {Lava, Stone, ScrubImpassible, BalancedForest, BalancedWater}


class Resource(metaclass=Meta):
    '''Materials corresponding to resources'''
    materials = {Forest, BalancedForest, Water, BalancedWater}

class Habitable(metaclass=Meta):
    '''Materials that agents cannot walk on'''
    materials = {Grass, Scrub, Forest}



from pdb import set_trace as T
from marlben.lib.colors import Tier


class Loadout:
    def __init__(self, chest=0, legs=0):
        self.chestplate = Chestplate(chest)
        self.platelegs = Platelegs(legs)

    @property
    def defense(self):
        return (self.chestplate.level + self.platelegs.level) // 2

    def packet(self):
        packet = {}

        packet['chestplate'] = self.chestplate.packet()
        packet['platelegs'] = self.platelegs.packet()

        return packet


class Armor:
    def __init__(self, level):
        self.level = level

    def packet(self):
        packet = {}

        packet['level'] = self.level
        packet['color'] = self.color.packet()

        return packet

    @property
    def color(self):
        if self.level == 0:
            return Tier.BLACK
        if self.level < 10:
            return Tier.WOOD
        elif self.level < 20:
            return Tier.BRONZE
        elif self.level < 40:
            return Tier.SILVER
        elif self.level < 60:
            return Tier.GOLD
        elif self.level < 80:
            return Tier.PLATINUM
        else:
            return Tier.DIAMOND


class Chestplate(Armor):
    pass


class Platelegs(Armor):
    pass



from typing import Callable

from marlben.core.spawn.spawn_system.position_samplers import RangePositionSampler
from marlben.core.spawn.spawn_system.skill_samplers import CustomSkillSampler
from marlben.io.action.attack import Range, Heal, Mage, Melee
from marlben.systems.achievement import Task
from marlben.core.map_generation.pregen_map_generator import PregeneratedMapGenerator
from marlben.config.base.config import NPCGroupConfig, PlayerGroupConfig, Config
from marlben.config.systems import NPC
from marlben.core.agent import Agent


class NpcKilledTask(Callable):
    __name__ = "Kill boss"

    def __init__(self, npc_group_id=0):
        self.dead_counter = 0
        self.npc_group_id = npc_group_id

    def __call__(self, realm, entity):
        # Amount of NPCs died at this turn
        self.dead_counter += len(realm.entity_group_manager.npc_groups[self.npc_group_id].dead)
        return self.dead_counter


class PlayerDiedTask(Callable):
    __name__ = "Die"

    def __call__(self, realm, entity):
        return 0 if entity.alive else 1


class BossGroupConfig(NPCGroupConfig):
    BANNED_ATTACK_STYLES = {Heal, Range, Mage}
    DANGER = 1.0  # Boss are aggressive by default
    NENT = 1
    SPAWN_ATTEMPTS_PER_ENT = 10

    def __init__(self, n_tanks, n_fighters, n_healers,
                 coordinates_sampler=RangePositionSampler(list(range(5, 8)), list(range(3, 8)))):
        super().__init__()
        self.SPAWN_COORDINATES_SAMPLER = coordinates_sampler
        self.SPAWN_SKILLS_SAMPLER = CustomSkillSampler(
            {"constitution": {"name": "const", "level": 125 * n_tanks + 250 * n_fighters + 1},
             "melee": {"name": "const", "level": 20 + 5 * n_healers}})
        self.DANGER = 1.0  # Boss are aggressive by default


class TankGroupConfig(PlayerGroupConfig):
    def __init__(self, n_ent=2, coordinates_sampler=RangePositionSampler(list(range(1, 4)), list(range(3, 7))),
                 skill_sampler=CustomSkillSampler({"constitution": {"name": "const", "level": 200},
                                                   "melee": {"name": "const", "level": 5}})):
        super().__init__()
        self.NENT = n_ent
        self.SPAWN_COORDINATES_SAMPLER = coordinates_sampler
        self.SPAWN_SKILLS_SAMPLER = skill_sampler

    SPAWN_ATTEMPTS_PER_ENT = 10
    BANNED_ATTACK_STYLES = [Range, Heal, Mage]
    AGENTS = [Agent]


class BossFightConfig(Config, NPC):
    NPC_GROUPS = [BossGroupConfig(2, 0, 0)]
    PLAYER_GROUPS = [TankGroupConfig(2)]
    TASKS = [Task(NpcKilledTask(0), 1, 10.), Task(PlayerDiedTask(), 1, -10.)]
    REGEN_HEALTH = 0.
    COMBAT_DEFENSE_WEIGHT = 1.

    TOP_LEFT_CORNER = (16, 17)
    TERRAIN_CENTER = 12
    MAP_HEIGHT = 10
    MAP_WIDTH = 12

    MAP_PREVIEW_DOWNSCALE = 1
    TERRAIN_LOG_INTERPOLATE_MIN = 0
    MAP_GENERATOR = PregeneratedMapGenerator


import marlben

"""
Similar to Team Deathmatch, agents are divided into two teams. However, only one of the teams, called the Predators, is capable of attacking,
while the Prey team's only goal is to survive. Also, the Predators team is smaller than the Prey team to balance the power.
This environment is similar to the environment of the same name from the MPE set [1]. However, there is still competition for resources between all agents.

Expected behavior:
With successful coordination, the Predators successfully capture all of the Preys and survive until the end of the episode.

Implementation details are located in config file

[1] Lowe R. et al. Multi-agent actor-critic for mixed cooperative-competitive environments // Advances in neural information processing systems. - 2017. - P. 6379–6390.
"""
class PredatorPrey(marlben.Env):
    pass



from pdb import set_trace as T

from marlben.systems.ai import behavior, utils
import random


def passive(realm, entity):
    behavior.update(entity)
    actions = {}

    behavior.meander(realm, actions, entity)

    return actions


def neutral(realm, entity):
    behavior.update(entity)
    actions = {}

    if not entity.attacker:
        if random.random() > 0.5:
            behavior.meander(realm, actions, entity)
    else:
        entity.target = entity.attacker
        behavior.hunt(realm, actions, entity)

    return actions


def hostile(realm, entity):
    behavior.update(entity)
    actions = {}

    # This is probably slow
    if not entity.target:
        entity.target = utils.closestTarget(entity, realm.map.tiles,
                                            rng=entity.vision)

    if not entity.target:
        behavior.meander(realm, actions, entity)
    else:
        behavior.hunt(realm, actions, entity)

    return actions



import marlben

"""
The specificity of this environment lies in the fact that at the beginning of the episode, only a part of the forest and water cells is visible to each agent,
while the rest of the forest and water cells are hidden and look like stone cells.
At the same time, different agents see different cells of the forest and the water. 
However, agents can still interact with a hidden cell and receive resources from it, thus making it important to spy on other agents routes

Expected behavior:
If the location of the forest and water cells visible to the agent is successful (that is, close to each other and to the agent), he may not look for other resource cells.
However, if the visible resource tiles are far apart, it may be easier for the agent to find other resource tiles on the map.
To do this, the agent can bypass the map until it accidentally stumbles upon resources.
A more sophisticated strategy is to monitor the behavior of other agents in an attempt to figure out the locations of the resources they see.
If opponents use this strategy, it can also be advantageous to sometimes move in arbitrary directions to confuse opponents. 
This is an element of adversarial communication in the environment, in what way it resembles, for example, Adversary from the MPE set [1]. 
Also, the environment can be useful for testing communication channels between agents for mutually beneficial information exchange.

Implementation details are located in config file

[1] Lowe R. et al. Multi-agent actor-critic for mixed cooperative-competitive environments // Advances in neural information processing systems. - 2017. - P. 6379–6390.
"""
class GatheringObscured(marlben.Env):
    pass



import marlben

from . import scale
from . import bases


class Medium(scale.Baseline, bases.Medium, marlben.config.AllGameSystems):
    # Load 1000 epoch pretrained model
    RESTORE = True
    RESTORE_ID = '870d'

    @property
    def SPAWN(self):
        return self.SPAWN_CONCURRENT

    pass


class Debug(scale.Debug, bases.Small, marlben.config.AllGameSystems):
    RESTORE = False

    TRAINING_ITERATIONS = 2

    SGD_MINIBATCH_SIZE = 100
    TRAIN_BATCH_SIZE = 400
    TRAIN_HORIZON = 200
    EVALUATION_HORIZON = 50

    HIDDEN = 2
    EMBED = 2



"""
Enables resources for agents
"""
class Resource:
    """Resource Game System"""

    @property  # Reserved flag
    def Resource(self):
        return True

    RESOURCE_BASE_RESOURCE = 10
    '''Initial level and capacity for Hunting + Fishing resource skills'''

    RESOURCE_FOREST_CAPACITY = 1
    '''Maximum number of harvests before a forest tile decays'''

    RESOURCE_FOREST_RESPAWN = 0.025
    '''Probability that a harvested forest tile will regenerate each tick'''

    RESOURCE_HARVEST_RESTORE_FRACTION = 1.0
    '''Fraction of maximum capacity restored upon collecting a resource'''

    RESOURCE_HEALTH_REGEN_THRESHOLD = 0.5
    '''Fraction of maximum resource capacity required to regen health'''

    RESOURCE_HEALTH_RESTORE_FRACTION = 0.1
    '''Fraction of health restored per tick when above half food+water'''

    REGEN_HEALTH = 1.
    RESOURCE_COOLDOWN = 1

"""
Enables attack action
"""
class Combat:
    '''Combat Game System'''

    @property  # Reserved flag
    def Combat(self):
        return True

    COMBAT_DICE_SIDES = 20
    '''Number of sides for combat dice

   Attacks can only hit opponents up to the attacker's level plus
   DICE_SIDES/2. Increasing this value makes attacks more accurate
   and allows lower level attackers to hit stronger opponents'''

    COMBAT_DEFENSE_WEIGHT = 0.3
    '''Fraction of defense that comes from the Defense skill'''

    COMBAT_MELEE_REACH = 1
    '''Reach of attacks using the Melee skill'''

    COMBAT_RANGE_REACH = 3
    '''Reach of attacks using the Range skill'''

    COMBAT_MAGE_REACH = 4
    '''Reach of attacks using the Mage skill'''

    COMBAT_HEAL_REACH = 2
    '''Reach of healing skill'''

    COMBAT_FREEZE_TIME = 3
    '''Number of ticks successful Mage attacks freeze a target'''

    STEALING_ENABLED = True
    '''Is attack allows to steal resources'''

"""
Enables skill progression for agents
"""
class Progression:
    '''Progression Game System'''

    @property  # Reserved flag
    def Progression(self):
        return True

    PROGRESSION_BASE_RESOURCE = 10
    '''Initial level and capacity for Hunting + Fishing resource skills'''

    PROGRESSION_BASE_XP_SCALE = 10
    '''Skill level progression speed as a multiplier of typical MMOs'''

    PROGRESSION_CONSTITUTION_XP_SCALE = 2
    '''Multiplier on top of XP_SCALE for the Constitution skill'''

    PROGRESSION_COMBAT_XP_SCALE = 4
    '''Multiplier on top of XP_SCALE for Combat skills'''

"""
Enables NPC spawns on the map
"""
class NPC(Combat):
    '''NPC & Equipment Game System'''

    @property  # Reserved flag
    def NPC(self):
        return True

    NPC_SPAWN_ATTEMPTS = 25
    '''Number of NPC spawn attempts per tick'''

    NPC_SPAWN_AGGRESSIVE = 0.80
    '''Percentage distance threshold from spawn for aggressive NPCs'''

    NPC_SPAWN_NEUTRAL = 0.50
    '''Percentage distance threshold from spawn for neutral NPCs'''

    NPC_SPAWN_PASSIVE = 0.00
    '''Percentage distance threshold from spawn for passive NPCs'''

    NPC_LEVEL_MIN = 1
    '''Minimum NPC level'''

    NPC_LEVEL_MAX = None
    '''Maximum NPC level'''

    NPC_LEVEL_SPREAD = None
    '''Level range for NPC spawns'''

"""
Enables sharing action for agents
"""
class Sharing(Resource):
    @property
    def Sharing(self):
        return True

    SHARING_DISTANCE = 3
    SHARE_MIN = 10
    SHARE_MAX = 10

"""
Enables building action for agents
"""
class Building:
    @property
    def Building(self):
        return True

"""
Enables planting action for agents
"""
class Planting:
    @property
    def Planting(self):
        return True

    PLANTING_COST = 8



# Various utilities for managing combat, including hit/damage

import numpy as np

from marlben.systems import skill as Skill


def level(skills):
    hp = skills.constitution.level
    defense = skills.defense.level
    melee = skills.melee.level
    ranged = skills.range.level
    mage = skills.mage.level

    base = 0.25*(defense + hp)
    final = np.floor(base + 0.5*max(melee, ranged, mage))
    return final

"""
Implements attack action, where entity attacks target for a fixed amount of damage
"""
def attack(entity, targ, skillFn, stealing_enabled=True):
    config = entity.config
    entitySkill = skillFn(entity)
    targetSkill = skillFn(targ)

    targetDefense = targ.skills.defense.level + targ.loadout.defense

    dmg = damage(entitySkill.__class__, entitySkill.level)
    if dmg > 0:
        die = config.COMBAT_DICE_SIDES
        roll = np.random.randint(1, die+1)
        dc = accuracy(config, entitySkill.level, targetSkill.level, targetDefense)
        crit = roll == die

        if not crit and roll < dc:
            dmg = 1

    # Note: damage can be negative!
    dmg = min(dmg, entity.resources.health.val)
    entity.applyDamage(
        dmg, entitySkill.__class__.__name__.lower(), stealing_enabled)
    targ.receiveDamage(entity, dmg, stealing_enabled)
    return dmg

# Compute maximum damage roll


def damage(skill, level):
    if skill == Skill.Melee:
        return np.floor(7 + level * 63 / 99)
    if skill == Skill.Range:
        return np.floor(3 + level * 32 / 99)
    if skill == Skill.Mage:
        return np.floor(1 + level * 24 / 99)
    if skill == Skill.Heal:
        return -np.floor(3 + level * 32 / 99)

# Compute maximum attack or defense roll (same formula)
# Max attack 198 - min def 1 = 197. Max 198 - max 198 = 0
# REMOVE FACTOR OF 2 FROM ATTACK AFTER IMPLEMENTING WEAPONS


def accuracy(config, entAtk, targAtk, targDef):
    alpha = config.COMBAT_DEFENSE_WEIGHT

    attack = entAtk
    defense = alpha*targDef + (1-alpha)*targAtk
    dc = defense - attack + config.COMBAT_DICE_SIDES//2

    return dc


def danger(config, pos):
    border = config.TERRAIN_BORDER
    center = config.TERRAIN_CENTER
    r, c = pos

    # Distance from border
    rDist = min(r - border, center + border - r - 1)
    cDist = min(c - border, center + border - c - 1)
    dist = min(rDist, cDist)
    norm = 2 * dist / center

    return norm



"""
Implements sharing action between two agents, where entity gives target a set amount of resources
"""
def share(entity, targ, resourceFn, amount):
    entity_resource = resourceFn(entity)
    targ_resource = resourceFn(targ)
    amount = min(entity_resource.val, amount,
                 targ_resource.max - targ_resource.val)
    entity_resource.decrement(amount)
    targ_resource.increment(amount)

    return amount



from os import path as osp
from typing import Callable

from marlben import Task
from marlben.config import Combat
from marlben.envs.gathering.config import GatheringConfig, GatheringConfigScripted

from marlben.scripted.environments.gathering import GatheringCombatAgent



PATH_TO_CUSTOM_MAPS = osp.dirname(__file__)

class AgentKilledTask(Callable):
    __name__ = "AgentKilled"
    def __call__(self, realm, entity):
        if entity.dead:
            return 0

        # Amount of Players died at this turn
        return sum([len(group.dead) for group in realm.entity_group_manager.player_groups])

class DeathmatchConfig(GatheringConfig, Combat):
    TASKS = [Task(AgentKilledTask(), 1, 10.)]
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = f'maps/deathmatch_{self.MAP_WIDTH}x{self.MAP_HEIGHT}'

class DeathmatchConfigScripted(GatheringConfigScripted, Combat):
    AGENT_TYPE = GatheringCombatAgent
    def __init__(self, n_groups, agents_per_group):
        super().__init__(n_groups, agents_per_group)
        self.PATH_MAPS = osp.join(PATH_TO_CUSTOM_MAPS, "test_maps")


import os

import numpy as np

from marlben import core
from marlben.lib import material


class Map:
    '''Map object representing a list of tiles

   Also tracks a sparse list of tile updates
   '''

    def __init__(self, config, realm):
        self.config = config

        self.tiles = np.zeros((config.MAP_HEIGHT + 2*config.TERRAIN_BORDER, config.MAP_WIDTH + 2*config.TERRAIN_BORDER),
                              dtype=object)

        for r in range(config.MAP_HEIGHT + 2*config.TERRAIN_BORDER):
            for c in range(config.MAP_WIDTH + 2*config.TERRAIN_BORDER):
                self.tiles[r, c] = core.Tile(config, realm, r, c)

    @property
    def packet(self):
        '''Packet of degenerate resource states'''
        missingResources = []
        for e in self.updateList:
            missingResources.append(e.pos)
        return missingResources

    @property
    def repr(self):
        '''Flat matrix of tile material indices'''
        return [[t.mat.index for t in row] for row in self.tiles]

    def reset(self, realm, idx):
        '''Reuse the current tile objects to load a new map'''
        config = self.config
        self.updateList = set()

        path_map_suffix = config.PATH_MAP_SUFFIX.format(idx)
        fPath = os.path.join(
            config.PATH_CWD, config.PATH_MAPS, path_map_suffix)

        try:
            map_file = np.load(fPath)
        except FileNotFoundError:
            print('Maps not found')
            raise

        materials = {mat.index: mat for mat in material.All}
        for r, row in enumerate(map_file):
            for c, (idx, vc, ac) in enumerate(row):
                mat = materials[idx]
                tile = self.tiles[r, c]
                tile.reset(mat, vc, ac, config)

    def step(self):
        '''Evaluate updatable tiles'''
        for e in self.updateList.copy():
            if e.static:
                self.updateList.remove(e)
            e.step()

    def harvest(self, r, c):
        '''Called by actions that harvest a resource tile'''
        self.updateList.add(self.tiles[r, c])
        return self.tiles[r, c].harvest()
